from contextlib import suppress
from datetime import datetime
from decimal import Decimal
from typing import TYPE_CHECKING

from django.db.models import Q
from wbcore.contrib.authentication.models.users import User
from wbcore.contrib.io.imports import ImportExportHandler
from wbcore.contrib.notifications.dispatch import send_notification

from wbportfolio.models import Portfolio

if TYPE_CHECKING:
    from wbportfolio.models.portfolio_cash_flow import DailyPortfolioCashFlow


class DailyPortfolioCashFlowImportHandler(ImportExportHandler):
    MODEL_APP_LABEL = "wbportfolio.DailyPortfolioCashFlow"

    def _deserialize(self, data):
        data["value_date"] = datetime.strptime(data["value_date"], "%Y-%m-%d").date()
        data["portfolio"] = Portfolio.all_objects.get(id=data["portfolio"])
        if "cash" in data:
            data["cash"] = Decimal(data["cash"])

        if "total_assets" in data:
            data["total_assets"] = Decimal(data["total_assets"])

        if "cash_flow_forecast" in data:
            data["cash_flow_forecast"] = Decimal(data["cash_flow_forecast"])

    def _get_instance(self, data, history=None, **kwargs):
        self.import_source.log += "\nGet Daily Portfolio Cash Flow Instance."
        self.import_source.log += f"\nParameter: Portfolio={data['portfolio']} Value Date={data['value_date']}"

        with suppress(self.model.DoesNotExist):
            return self.model.objects.get(portfolio=data["portfolio"], value_date=data["value_date"])

    def _post_processing_created_object(self, _object: "DailyPortfolioCashFlow"):
        if not _object.portfolio.daily_cashflows.filter(value_date__gt=_object.value_date).exists():
            if _object.proposed_rebalancing != Decimal(0):
                color = "red" if _object.proposed_rebalancing < 0 else "green"
                for user in User.objects.filter(
                    Q(user_permissions__codename="view_dailyportfoliocashflow")
                    | Q(groups__permissions__codename="view_dailyportfoliocashflow")
                ).distinct():
                    send_notification(
                        code="wbportfolio.dailyportfoliocashflow.notify_rebalance",
                        title=f"Proposed Rebalancing in the portfolio: {_object.portfolio}",
                        body=f"The workbench proposes to rebalance the portfolio {_object.portfolio} by <span style='color:{color}'>{_object.proposed_rebalancing:+,.2f}</span>",
                        user=user,
                        reverse_name="wbportfolio:portfoliocashflow-detail",
                        reverse_args=[_object.pk],
                    )



from wbcore.menus import ItemPermission, MenuItem

PRODUCTCUSTOMER_MENUITEM = MenuItem(
    label="Product Lists",
    endpoint="wbportfolio:productcustomer-list",
    permission=ItemPermission(method=lambda request: not request.user.is_internal and not request.user.is_superuser),
)
PRODUCT_MENUITEM = MenuItem(
    label="Products",
    endpoint="wbportfolio:product-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbportfolio.view_product"]
    ),
)



from datetime import datetime

import numpy as np
import pandas as pd
from django.db.models import F, OuterRef, Subquery
from rest_framework.exceptions import ParseError
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.contrib.pandas import fields as pf
from wbcore.contrib.permission.permissions import InternalUserPermissionMixin
from wbcore.serializers import decorator
from wbcore.utils.strings import format_number
from wbfdm.models import Classification, ClassificationGroup, Instrument

from wbportfolio.filters import (
    AggregatedAssetPositionLiquidityFilter,
    AssetPositionPandasFilter,
)
from wbportfolio.filters.positions import GroupbyChoice
from wbportfolio.models import AssetPosition, Portfolio

from ..constants import EQUITY_TYPE_KEYS
from .configs import (
    AggregatedAssetPositionLiquidityDisplayConfig,
    AggregatedAssetPositionLiquidityEndpointConfig,
    AggregatedAssetPositionLiquidityTitleConfig,
    AssetPositionPandasDisplayConfig,
    AssetPositionPandasEndpointConfig,
    AssetPositionPandasTitleConfig,
)
from .mixins import UserPortfolioRequestPermissionMixin


class AssetPositionPandasView(
    UserPortfolioRequestPermissionMixin, InternalUserPermissionMixin, ExportPandasAPIViewSet
):
    IDENTIFIER = "wbportfolio:equityposition"

    queryset = AssetPosition.objects.all()
    filterset_class = AssetPositionPandasFilter

    pandas_fields = pf.PandasFields(
        fields=(
            pf.PKField(key="id", label="ID"),
            pf.CharField(key="title", label="Title"),
            pf.FloatField(key="performance_total", label="Total Performance", precision=2, percent=True),
            pf.FloatField(key="performance_forex", label="Forex Performance", precision=2, percent=True),
            pf.FloatField(key="contribution_total", label="Total Contribution", precision=2, percent=True),
            pf.FloatField(key="contribution_forex", label="Forex Contribution", precision=2, percent=True),
            pf.FloatField(key="allocation_start", label="Value Start", precision=2, percent=True),
            pf.FloatField(key="allocation_end", label="Value End", precision=2, percent=True),
            # pf.FloatField(key="price_start", label="Price Start", precision=1),
            # pf.FloatField(key="price_end", label="Price End", precision=1),
            pf.FloatField(
                key="total_value_start",
                label="Total Value Start",
                precision=1,
                decorators=(decorator(decorator_type="text", position="left", value="$"),),
            ),
            pf.FloatField(
                key="total_value_end",
                label="Total Value End",
                precision=1,
                decorators=(decorator(decorator_type="text", position="left", value="$"),),
            ),
            pf.FloatField(key="market_share", label="Market Shares", precision=4, percent=True),
        )
    )
    display_config_class = AssetPositionPandasDisplayConfig
    title_config_class = AssetPositionPandasTitleConfig
    endpoint_config_class = AssetPositionPandasEndpointConfig

    search_fields = ["title"]
    ordering_fields = [
        "title",
        "allocation_start",
        "total_value_start",
        "total_value_end",
        "allocation_end",
        "performance_total",
        "contribution_total",
        "performance_forex",
        "contribution_forex",
        "market_share",
    ]

    def get_aggregates(self, request, df):
        if df.empty:
            return {}
        return {
            "allocation_start": {"Σ": format_number(df["allocation_start"].sum())},
            "allocation_end": {"Σ": format_number(df["allocation_end"].sum())},
            "total_value_start": {"Σ": format_number(df["total_value_start"].sum())},
            "total_value_end": {"Σ": format_number(df["total_value_end"].sum())},
            "contribution_total": {"Σ": format_number(df["contribution_total"].sum())},
            "contribution_forex": {"Σ": format_number(df["contribution_forex"].sum())},
        }

    def get_queryset(self):
        if self.is_analyst:
            return (
                super()
                .get_queryset()
                .filter(
                    shares__isnull=False, is_invested=True
                )  # We remove assets position coming from non real portfolio (e.g. index)
                .annotate(
                    currency_fx_rate_usd=CurrencyFXRates.get_fx_rates_subquery(
                        "date", currency="currency", lookup_expr="exact"
                    )
                )
            )
        return AssetPosition.objects.none()

    def get_dataframe(self, request, queryset, **kwargs):
        df = pd.DataFrame()
        if queryset.exists() and (
            primary_classification_group := ClassificationGroup.objects.filter(is_primary=True).first()
        ):
            groupby = GroupbyChoice(request.GET.get("group_by", GroupbyChoice.UNDERLYING_INSTRUMENT))
            groupby_classification_height = int(request.GET.get("groupby_classification_height", "0"))

            groupby_id = GroupbyChoice.get_id(groupby.name)
            if groupby == GroupbyChoice.PRIMARY_CLASSIFICATION:
                queryset = queryset.annotate_classification_for_group(
                    primary_classification_group, classification_height=groupby_classification_height, unique=True
                ).annotate(
                    classification_id=F("classifications"),
                    classification_title=Subquery(
                        Classification.objects.filter(id=OuterRef("classification_id")).values("name")[:1]
                    ),
                )
            elif groupby == GroupbyChoice.PREFERRED_CLASSIFICATION:
                queryset = queryset.annotate_preferred_classification_for_group(
                    primary_classification_group, classification_height=groupby_classification_height
                )
            df = Portfolio.get_contribution_df(
                queryset.values_list("date", "price", "currency_fx_rate_usd", groupby_id, "weighting"),
                need_normalize=True,
            )
            df = df.rename(
                columns={
                    "group_key": "id",
                }
            )
            df["title"] = df["id"].map(groupby.get_repr(df["id"]))
            if groupby == GroupbyChoice.UNDERLYING_INSTRUMENT.value:
                df_market_shares = pd.DataFrame(
                    queryset.values("market_share", "date", "underlying_instrument")
                ).pivot_table(
                    index="underlying_instrument",
                    columns=["date"],
                    values="market_share",
                    fill_value=None,
                    aggfunc="sum",
                )
                df = df.set_index("id")
                df["market_share"] = df_market_shares.iloc[:, -1]
                df = df.reset_index()

        return df

    def manipulate_dataframe(self, df):
        if "titl" in df.columns:
            df["title"] = df["title"].fillna("N/A").astype(str)
        # df[df["title"].isnull()] = "N/A"
        return df.fillna(0)


class AggregatedAssetPositionLiquidityPandasView(InternalUserPermissionMixin, ExportPandasAPIViewSet):
    IDENTIFIER = "wbportfolio:aggregatedassetpositionliquidity"
    LIST_DOCUMENTATION = "wbportfolio/markdown/documentation/aggregate_asset_position_liquidity.md"
    queryset = AssetPosition.objects.all()

    def get_pandas_fields(self, request):
        historic_date = self.request.GET.get("historic_date", None)
        compared_date = self.request.GET.get("compared_date", None)
        fields = [
            pf.PKField(key="id", label="ID"),
            pf.CharField(key="name", label="Name"),
            pf.CharField(
                key="shares_first_date", label=f"Total Shares - {historic_date}" if historic_date else "None"
            ),
            pf.CharField(
                key="volume_50d_first_date", label=f"Mean Volume - {historic_date}" if historic_date else "None"
            ),
            pf.CharField(
                key="liquidity_first_date", label=f"Days To Liquidate - {historic_date}" if historic_date else "None"
            ),
            pf.CharField(
                key="pct_aum_first_date", label=f"Percent AUM - {historic_date}" if historic_date else "None"
            ),
            pf.CharField(
                key="liquidity_second_date", label=f"Days To Liquidate - {compared_date}" if compared_date else "None"
            ),
            pf.CharField(
                key="pct_aum_second_date", label=f"Percent AUM - {compared_date}" if compared_date else "None"
            ),
        ]
        return pf.PandasFields(fields=list(fields))

    title_config_class = AggregatedAssetPositionLiquidityTitleConfig
    endpoint_config_class = AggregatedAssetPositionLiquidityEndpointConfig
    display_config_class = AggregatedAssetPositionLiquidityDisplayConfig

    filterset_class = AggregatedAssetPositionLiquidityFilter

    def get_dataframe(self, request, queryset, **kwargs):
        if "historic_date" not in request.GET or "compared_date" not in request.GET:
            raise ParseError()
        historic_date = datetime.strptime(request.GET["historic_date"], "%Y-%m-%d").date()
        compared_date = datetime.strptime(request.GET["compared_date"], "%Y-%m-%d").date()

        # Take the liquidity query for the two dates.
        qs_assets = queryset.filter(
            underlying_instrument__instrument_type__key__in=EQUITY_TYPE_KEYS,
            date__in=[historic_date, compared_date],
        ).values(
            "date",
            "underlying_instrument",
            "underlying_instrument__name",
            "portfolio",
            "shares",
            "liquidity",
            "volume_50d",
            "total_value_fx_usd",
        )

        if not qs_assets.exists():
            return pd.DataFrame()
        df_assets = pd.DataFrame(qs_assets).set_index(["date", "underlying_instrument", "portfolio"])
        # We aggregate by date and underlying instrument to drop the portfolio level.
        df_assets = df_assets.groupby(["date", "underlying_instrument"]).agg(
            {
                "underlying_instrument__name": "first",
                "shares": "sum",
                "liquidity": "sum",
                "total_value_fx_usd": "sum",
                "volume_50d": "first",
            }
        )
        df_assets = df_assets.reset_index("underlying_instrument").set_index("underlying_instrument", append=True)
        df_assets["pct_aum"] = df_assets.total_value_fx_usd / df_assets.total_value_fx_usd.groupby("date").sum() * 100
        df_assets = df_assets.reindex(columns=["shares", "volume_50d", "liquidity", "pct_aum"]).unstack("date")

        # Make a distinction between first & second date for the columns.
        df_assets.columns = pd.Index(
            [(f"{a}_first_date" if b == historic_date else f"{a}_second_date") for a, b in df_assets.columns]
        )

        # Sort by days to liquidate.
        if "liquidity_first_date" in df_assets.columns:
            df_assets.sort_values(by="liquidity_first_date", ascending=False, inplace=True)
        else:  # if first one does not exist, at least second one exists.
            df_assets.sort_values(by="liquidity_second_date", ascending=False, inplace=True)

        # Cut dataframe for days to liquidate smaller than a specific value.
        if bigger_than_x := request.GET.get("bigger_than_x"):
            if "liquidity_first_date" in df_assets.columns:
                df_assets = df_assets[df_assets["liquidity_first_date"] >= float(bigger_than_x)]
            else:
                df_assets = df_assets[df_assets["liquidity_second_date"] >= float(bigger_than_x)]
        return df_assets

    def manipulate_dataframe(self, df):
        df = df.astype(float).round(2)
        # Make empty columns if they do not exist.
        if df.filter(like="first_date").empty:
            df[["liquidity_first_date", "pct_aum_first_date", "shares_first_date", "volume_50d_first_date"]] = np.nan
        if df.filter(like="second_date").empty:
            df[["liquidity_second_date", "pct_aum_second_date"]] = np.nan

        # Put % for percent columns.
        df.loc[:, df.filter(like="pct_aum").columns] = df.filter(like="pct_aum").applymap(
            lambda x: str(x) + "%" if not np.isnan(x) else None
        )

        # Round big numbers columns and put an apostrophe.
        df.loc[:, ["shares_first_date", "volume_50d_first_date"]] = (
            df.loc[:, ["shares_first_date", "volume_50d_first_date"]]
            .round(0)
            .applymap(lambda x: f"{int(x):,}" if not np.isnan(x) else x)
        )
        df = df.reset_index()
        if not df.empty:
            df["name"] = df["underlying_instrument"].map(
                dict(
                    Instrument.objects.filter(id__in=df["underlying_instrument"].unique()).values_list(
                        "id", "computed_str"
                    )
                )
            )
        return df.rename(columns={"underlying_instrument": "id"})



import logging
from datetime import date

from celery import shared_task
from django.db.models import F, QuerySet, Window
from django.db.models.functions import RowNumber
from tqdm import tqdm
from wbcore.workers import Queue
from wbfdm.models import Instrument

from wbportfolio.models import AssetPosition, Order, OrderProposal, Portfolio
from wbportfolio.models.asset import adjust_assets
from wbportfolio.models.orders.orders import adjust_orders

logger = logging.getLogger("pms")


def adjust_quote(
    old_quote: Instrument,
    new_quote: Instrument,
    adjust_after: date | None = None,
    only_portfolios: QuerySet[Portfolio] | None = None,
    debug: bool = False,
):
    if old_quote.currency != new_quote.currency:
        raise ValueError("cannot safely switch quotes that are not of the same currency")
    assets_to_change = AssetPosition.objects.filter(underlying_quote=old_quote)
    orders_to_change = Order.objects.filter(underlying_instrument=old_quote)
    new_quote.import_prices()
    if adjust_after:
        assets_to_change = assets_to_change.filter(date__gt=adjust_after)
        orders_to_change = orders_to_change.filter(value_date__gt=adjust_after)
    if only_portfolios is not None:
        assets_to_change = assets_to_change.filter(portfolio__in=only_portfolios)
        orders_to_change = orders_to_change.filter(order_proposal__portfolio__in=only_portfolios)
    if debug:
        assets_to_change = tqdm(assets_to_change, total=assets_to_change.count())
        orders_to_change = tqdm(orders_to_change, total=orders_to_change.count())

    # gather the list of order proposal to replay (if the quote led to missing position, we want to replay it to correct automatically the issue)
    latest_orders = orders_to_change.annotate(
        row_number=Window(
            expression=RowNumber(), partition_by=[F("order_proposal__portfolio")], order_by=F("value_date").desc()
        )
    ).filter(row_number=1)
    order_proposals_to_replay = OrderProposal.objects.filter(
        portfolio__is_manageable=True, id__in=latest_orders.values("order_proposal")
    )
    logger.info("adjusting asset positions...")
    # Adjust assets to the new quote
    adjust_assets(assets_to_change, new_quote)
    logger.info("adjusting orders...")
    # Adjust orders to the new quote
    adjust_orders(orders_to_change, new_quote)

    # replay latest order proposal
    for op in order_proposals_to_replay:
        op.replay(reapply_order_proposal=True)


@shared_task(queue=Queue.BACKGROUND.value)
def adjust_quote_as_task(
    old_quote_id: int, new_quote_id: int, adjust_after: date | None = None, only_portfolio_ids: list[int] | None = None
):
    old_quote = Instrument.objects.get(id=old_quote_id)
    new_quote = Instrument.objects.get(id=new_quote_id)
    only_portfolios = Portfolio.objects.filter(id__in=only_portfolio_ids) if only_portfolio_ids else None
    adjust_quote(old_quote, new_quote, adjust_after=adjust_after, only_portfolios=only_portfolios)



# Generated by Django 5.0.14 on 2025-07-17 14:53

import django.db.models.deletion
import django_fsm
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('currency', '0001_initial'),
        ('io', '0008_importsource_resource_kwargs'),
        ('wbfdm', '0031_exchange_apply_round_lot_size_and_more'),
        ('directory', '0013_alter_clientmanagerrelationship_options'),
        ('wbportfolio', '0081_alter_trade_drift_factor'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='TradeProposal',
            new_name='OrderProposal',
        ),
        migrations.AlterField(
            model_name='orderproposal',
            name='creator',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT,
                                    related_name='order_proposals', to='directory.person', verbose_name='Owner'),
        ),
        migrations.AlterField(
            model_name='orderproposal',
            name='portfolio',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='order_proposals',
                                    to='wbportfolio.portfolio', verbose_name='Portfolio'),
        ),
        migrations.AlterField(
            model_name='orderproposal',
            name='rebalancing_model',
            field=models.ForeignKey(blank=True, help_text='Rebalancing Model that generates the target portfolio',
                                    null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    related_name='order_proposals', to='wbportfolio.rebalancingmodel',
                                    verbose_name='Rebalancing Model'),
        ),
        migrations.AlterModelOptions(
            name='orderproposal',
            options={'verbose_name': 'Order Proposal', 'verbose_name_plural': 'Order Proposals'},
        ),
        migrations.RemoveConstraint(
            model_name='orderproposal',
            name='unique_trade_proposal',
        ),
        migrations.RemoveConstraint(
            model_name='trade',
            name='unique_manual_trade',
        ),
        migrations.RenameField(
            model_name='rebalancer',
            old_name='approve_trade_proposal_automatically',
            new_name='approve_order_proposal_automatically',
        ),
        migrations.RenameField(
            model_name='trade',
            old_name='trade_proposal',
            new_name='order_proposal',
        ),
        migrations.AlterField(
            model_name='portfolio',
            name='is_manageable',
            field=models.BooleanField(default=False,
                                      help_text='True if the portfolio can be manually modified (e.g. Order Proposal be submitted or total weight recomputed)'),
        ),
        migrations.AddConstraint(
            model_name='orderproposal',
            constraint=models.UniqueConstraint(fields=('portfolio', 'trade_date'), name='unique_order_proposal'),
        ),
        migrations.AddConstraint(
            model_name='trade',
            constraint=models.UniqueConstraint(condition=models.Q(('order_proposal__isnull', False)),
                                               fields=('portfolio', 'transaction_date', 'underlying_instrument'),
                                               name='unique_manual_trade'),
        ),
        migrations.AlterField(
            model_name='rebalancer',
            name='approve_order_proposal_automatically',
            field=models.BooleanField(default=False, verbose_name='Apply Order Proposal Automatically'),
        ),
        migrations.AlterField(
            model_name='trade',
            name='order_proposal',
            field=models.ForeignKey(blank=True, help_text='The Order Proposal this trade is coming from', null=True,
                                    on_delete=django.db.models.deletion.CASCADE, related_name='trades',
                                    to='wbportfolio.orderproposal'),
        ),
    ]



# Import necessary modules
from datetime import date, timedelta
from decimal import Decimal
from unittest.mock import MagicMock, PropertyMock, call, patch

import pytest
from django.db.models import Sum
from faker import Faker
from pandas._libs.tslibs.offsets import BDay, BusinessMonthEnd

from wbportfolio.models import Order, OrderProposal, Portfolio, RebalancingModel
from wbportfolio.order_routing import ExecutionInstruction, ExecutionStatus, RoutingException
from wbportfolio.pms.typing import Order as OrderDTO
from wbportfolio.pms.typing import Portfolio as PortfolioDTO
from wbportfolio.pms.typing import Position

fake = Faker()


@pytest.fixture
def mock_adapter():
    adapter = MagicMock()
    return adapter


# Mark tests to use Django's database
@pytest.mark.django_db
class TestOrderProposal:
    def test_init(self, order_proposal):
        assert order_proposal.id is not None

    # Test that the checked object is correctly set to the portfolio
    def test_checked_object(self, order_proposal):
        """
        Verify that the checked object is the portfolio associated with the order proposal.
        """
        assert order_proposal.checked_object == order_proposal.portfolio

    # Test that the evaluation date matches the trade date
    def test_check_evaluation_date(self, order_proposal):
        """
        Ensure the evaluation date is the same as the trade date.
        """
        assert order_proposal.check_evaluation_date == order_proposal.trade_date

    # Test the validated trading service functionality
    def test_validated_trading_service(
        self, order_proposal, asset_position_factory, instrument_price_factory, instrument_factory, order_factory
    ):
        """
        Validate that the effective and target portfolios are correctly calculated.
        """
        effective_date = (order_proposal.trade_date - BDay(1)).date()

        i1 = instrument_factory.create()
        i2 = instrument_factory.create()

        p10 = instrument_price_factory.create(instrument=i1, date=effective_date)
        p11 = instrument_price_factory.create(instrument=i1, date=order_proposal.trade_date)

        p20 = instrument_price_factory.create(instrument=i2, date=effective_date)
        p21 = instrument_price_factory.create(instrument=i2, date=order_proposal.trade_date)

        # Create asset positions for testing
        a1 = asset_position_factory.create(
            portfolio=order_proposal.portfolio, date=effective_date, weighting=Decimal("0.3"), underlying_instrument=i1
        )
        a2 = asset_position_factory.create(
            portfolio=order_proposal.portfolio, date=effective_date, weighting=Decimal("0.7"), underlying_instrument=i2
        )
        r1 = p11.net_value / p10.net_value - Decimal("1")
        r2 = p21.net_value / p20.net_value - Decimal("1")
        p_return = a1.weighting * (Decimal("1") + r1) + a2.weighting * (Decimal("1") + r2)
        order_proposal.total_effective_portfolio_contribution = p_return
        order_proposal.save()

        # Create orders for testing
        o1 = order_factory.create(
            order_proposal=order_proposal,
            weighting=Decimal("0.05"),
            portfolio=order_proposal.portfolio,
            underlying_instrument=i1,
        )
        o2 = order_factory.create(
            order_proposal=order_proposal,
            weighting=Decimal("-0.05"),
            portfolio=order_proposal.portfolio,
            underlying_instrument=i2,
        )

        # Get the validated trading service
        trades = order_proposal.get_trades_batch().trades_map
        t1 = trades[a1.underlying_quote.id]
        t2 = trades[a2.underlying_quote.id]

        # Assert effective and target portfolios are as expected
        assert t1.effective_weight == pytest.approx(
            a1.weighting * ((r1 + Decimal("1")) / p_return), abs=Decimal("1e-8")
        )
        assert t2.effective_weight == pytest.approx(
            a2.weighting * ((r2 + Decimal("1")) / p_return), abs=Decimal("1e-8")
        )
        assert t1.target_weight == pytest.approx(
            a1.weighting * ((r1 + Decimal("1")) / p_return) + o1.weighting, abs=Decimal("1e-8")
        )
        assert t2.target_weight == pytest.approx(
            a2.weighting * ((r2 + Decimal("1")) / p_return) + o2.weighting, abs=Decimal("1e-8")
        )

    # Test the calculation of the last effective date
    def test_last_effective_date(self, order_proposal, asset_position_factory):
        """
        Verify the last effective date is correctly determined based on asset positions.
        """
        # Without any positions, it should be the day before the trade date
        assert order_proposal.last_effective_date == (order_proposal.trade_date - BDay(1)).date(), (
            "Last effective date without position should be t-1"
        )

        # Create an asset position before the trade date
        a1 = asset_position_factory.create(
            portfolio=order_proposal.portfolio, date=(order_proposal.trade_date - BDay(5)).date()
        )
        a_noise = asset_position_factory.create(portfolio=order_proposal.portfolio, date=order_proposal.trade_date)  # noqa

        # The last effective date should still be the day before the trade date due to caching
        assert order_proposal.last_effective_date == (order_proposal.trade_date - BDay(1)).date(), (
            "last effective date is cached, so it won't change as is"
        )

        # Reset the cache property to recalculate
        del order_proposal.last_effective_date

        # Now it should be the date of the latest position before the trade date
        assert order_proposal.last_effective_date == a1.date, (
            "last effective date is the latest position strictly lower than trade date"
        )

    # Test finding the previous order proposal
    def test_previous_order_proposal(self, order_proposal_factory):
        """
        Ensure the previous order proposal is correctly identified as the last approved proposal before the current one.
        """
        tp = order_proposal_factory.create()
        tp_previous_submit = order_proposal_factory.create(  # noqa
            portfolio=tp.portfolio, status=OrderProposal.Status.PENDING, trade_date=(tp.trade_date - BDay(1)).date()
        )
        tp_previous_approve = order_proposal_factory.create(
            portfolio=tp.portfolio, status=OrderProposal.Status.CONFIRMED, trade_date=(tp.trade_date - BDay(2)).date()
        )
        tp_next_approve = order_proposal_factory.create(  # noqa
            portfolio=tp.portfolio, status=OrderProposal.Status.CONFIRMED, trade_date=(tp.trade_date + BDay(1)).date()
        )

        # The previous valid order proposal should be the approved one strictly before the current proposal
        assert tp.previous_order_proposal == tp_previous_approve, (
            "the previous valid order proposal is the strictly before and approved order proposal"
        )

    # Test finding the next order proposal
    def test_next_order_proposal(self, order_proposal_factory):
        """
        Verify the next order proposal is correctly identified as the first approved proposal after the current one.
        """
        tp = order_proposal_factory.create()
        tp_previous_approve = order_proposal_factory.create(  # noqa
            portfolio=tp.portfolio, status=OrderProposal.Status.CONFIRMED, trade_date=(tp.trade_date - BDay(1)).date()
        )
        tp_next_submit = order_proposal_factory.create(  # noqa
            portfolio=tp.portfolio, status=OrderProposal.Status.PENDING, trade_date=(tp.trade_date + BDay(1)).date()
        )
        tp_next_approve = order_proposal_factory.create(
            portfolio=tp.portfolio, status=OrderProposal.Status.CONFIRMED, trade_date=(tp.trade_date + BDay(2)).date()
        )

        # The next valid order proposal should be the approved one strictly after the current proposal
        assert tp.next_order_proposal == tp_next_approve, (
            "the next valid order proposal is the strictly after and approved order proposal"
        )

    # Test getting the default target portfolio
    def test_get_default_target_portfolio(self, order_proposal, asset_position_factory):
        """
        Ensure the default target portfolio is set to the effective portfolio from the day before the trade date.
        """
        effective_date = (order_proposal.trade_date - BDay(1)).date()

        # Create asset positions for testing
        a1 = asset_position_factory.create(
            portfolio=order_proposal.portfolio, date=effective_date, weighting=Decimal("0.3")
        )
        a2 = asset_position_factory.create(
            portfolio=order_proposal.portfolio, date=effective_date, weighting=Decimal("0.7")
        )
        asset_position_factory.create(portfolio=order_proposal.portfolio, date=order_proposal.trade_date)  # noise

        # The default target portfolio should match the effective portfolio
        assert order_proposal.get_default_target_portfolio().to_dict() == {
            a1.underlying_quote.id: a1.weighting,
            a2.underlying_quote.id: a2.weighting,
        }

    # Test getting the default target portfolio with a rebalancing model
    @patch.object(RebalancingModel, "get_target_portfolio")
    def test_get_default_target_portfolio_with_rebalancer_model(self, mock_fct, order_proposal, rebalancer_factory):
        """
        Verify that the target portfolio is correctly obtained from a rebalancing model.
        """
        # Expected target portfolio from the rebalancing model
        expected_target_portfolio = PortfolioDTO(
            positions=(Position(underlying_instrument=1, weighting=Decimal(1), date=order_proposal.trade_date),)
        )
        mock_fct.return_value = expected_target_portfolio

        # Create a rebalancer for testing
        rebalancer = rebalancer_factory.create(
            portfolio=order_proposal.portfolio, parameters={"rebalancer_parameter": "A"}
        )
        order_proposal.rebalancing_model = rebalancer.rebalancing_model
        order_proposal.save()

        # Additional keyword arguments for the rebalancing model
        extra_kwargs = {"test": "test"}

        # Combine rebalancer parameters with extra keyword arguments
        expected_kwargs = rebalancer.parameters
        expected_kwargs.update(extra_kwargs)

        # Assert the target portfolio matches the expected output from the rebalancing model
        assert order_proposal.get_default_target_portfolio(**extra_kwargs) == expected_target_portfolio, (
            "We expect the target portfolio to be whatever is returned by the rebalancer model"
        )
        mock_fct.assert_called_once_with(
            order_proposal.portfolio, order_proposal.trade_date, order_proposal.last_effective_date, **expected_kwargs
        )

    # Test normalizing orders
    def test_normalize_orders(self, order_proposal, order_factory):
        """
        Ensure orders are normalized to sum up to 1, handling quantization errors.
        """
        # Create orders for testing
        t1 = order_factory.create(
            order_proposal=order_proposal,
            portfolio=order_proposal.portfolio,
            weighting=Decimal(0.05),
        )
        t2 = order_factory.create(
            order_proposal=order_proposal,
            portfolio=order_proposal.portfolio,
            weighting=Decimal(0.22),
        )
        t3 = order_factory.create(
            order_proposal=order_proposal,
            portfolio=order_proposal.portfolio,
            weighting=Decimal(0.14),
        )

        # Normalize orders
        order_proposal.normalize_orders(Decimal("0.18"))

        # Refresh orders from the database
        t1.refresh_from_db()
        t2.refresh_from_db()
        t3.refresh_from_db()
        cash = order_proposal.orders.get(underlying_instrument__is_cash=True)

        # Expected normalized weights
        assert t1.weighting == Decimal("0.10")
        assert t2.weighting == Decimal("0.44")
        assert t3.weighting == Decimal("0.28")
        assert cash.weighting == Decimal("0.18")

    # Test resetting orders
    def test_reset_orders(
        self, order_proposal, instrument_factory, cash_factory, instrument_price_factory, asset_position_factory
    ):
        """
        Verify orders are correctly reset based on effective and target portfolios.
        """
        cash = cash_factory.create()
        effective_date = order_proposal.last_effective_date

        # Create instruments for testing
        i1 = instrument_factory.create(currency=order_proposal.portfolio.currency)
        i2 = instrument_factory.create(currency=order_proposal.portfolio.currency)
        i3 = instrument_factory.create(currency=order_proposal.portfolio.currency)
        # Build initial effective portfolio constituting only from two positions of i1 and i2
        asset_position_factory.create(
            portfolio=order_proposal.portfolio, date=effective_date, underlying_instrument=i1, weighting=Decimal("0.7")
        )
        asset_position_factory.create(
            portfolio=order_proposal.portfolio, date=effective_date, underlying_instrument=i2, weighting=Decimal("0.3")
        )
        p1 = instrument_price_factory.create(instrument=i1, date=effective_date)
        p2 = instrument_price_factory.create(instrument=i2, date=effective_date)
        p3 = instrument_price_factory.create(instrument=i3, date=effective_date)

        # build the target portfolio
        target_portfolio = PortfolioDTO(
            [
                Position(
                    underlying_instrument=i2.id,
                    date=order_proposal.trade_date,
                    weighting=Decimal("0.4"),
                    price=float(p2.net_value),
                ),
                Position(
                    underlying_instrument=i3.id,
                    date=order_proposal.trade_date,
                    weighting=Decimal("0.6"),
                    price=float(p3.net_value),
                ),
            ]
        )

        # Reset orders
        order_proposal.reset_orders(target_portfolio=target_portfolio)

        # Get orders for each instrument
        t1 = order_proposal.orders.get(underlying_instrument=i1)
        t2 = order_proposal.orders.get(underlying_instrument=i2)
        t3 = order_proposal.orders.get(underlying_instrument=i3)

        # Assert trade weights are correctly reset
        assert t1.weighting == Decimal("-0.7")
        assert t2.weighting == Decimal("0.1")
        assert t3.weighting == Decimal("0.6")

        # build the target portfolio
        new_target_portfolio = PortfolioDTO(
            [
                Position(
                    underlying_instrument=i1.id,
                    date=order_proposal.trade_date,
                    weighting=Decimal("0.2"),
                    price=float(p1.net_value),
                ),
                Position(
                    underlying_instrument=i2.id,
                    date=order_proposal.trade_date,
                    weighting=Decimal("0.3"),
                    price=float(p2.net_value),
                ),
                Position(
                    underlying_instrument=i3.id,
                    date=order_proposal.trade_date,
                    weighting=Decimal("0.5"),
                    price=float(p3.net_value),
                ),
            ]
        )

        order_proposal.reset_orders(target_portfolio=new_target_portfolio)
        # Refetch the orders for each instrument
        t1.refresh_from_db()
        t2.refresh_from_db()
        t3.refresh_from_db()
        # Assert existing trade weights are correctly updated
        assert t1.weighting == Decimal("-0.5")
        assert t2.weighting == Decimal("0")
        assert t3.weighting == Decimal("0.5")

        # assert cash position creates a proper order
        # build the target portfolio
        target_portfolio_with_cash = PortfolioDTO(
            [
                Position(
                    underlying_instrument=i1.id,
                    date=order_proposal.trade_date,
                    weighting=Decimal("0.5"),
                    price=float(p1.net_value),
                ),
                Position(
                    underlying_instrument=cash.id,
                    date=order_proposal.trade_date,
                    weighting=Decimal("0.5"),
                    price=1.0,
                ),
            ]
        )
        order_proposal.reset_orders(target_portfolio=target_portfolio_with_cash)

        # Assert existing trade weights are correctly updated
        assert order_proposal.orders.get(underlying_instrument=i1).weighting == Decimal("-0.2")
        assert order_proposal.orders.get(underlying_instrument=i2).weighting == Decimal("-0.3")
        assert order_proposal.orders.get(underlying_instrument=cash).weighting == Decimal("0.5")

    def test_reset_orders_remove_invalid_orders(self, order_proposal, order_factory):
        # create a invalid trade and its price
        invalid_trade = order_factory.create(order_proposal=order_proposal, weighting=Decimal(0))

        # create a valid trade and its price
        valid_trade = order_factory.create(order_proposal=order_proposal, weighting=Decimal(1))
        order_proposal.reset_orders()
        assert order_proposal.orders.get(underlying_instrument=valid_trade.underlying_instrument).weighting == Decimal(
            "1"
        )
        with pytest.raises(Order.DoesNotExist):
            order_proposal.orders.get(underlying_instrument=invalid_trade.underlying_instrument)

    # Test replaying order proposals
    @patch.object(Portfolio, "drift_weights")
    def test_replay(self, mock_fct, order_proposal_factory):
        """
        Ensure replaying order proposals correctly calls drift_weights for each period.
        """
        mock_fct.return_value = iter([])

        # Create approved order proposals for testing
        tp0 = order_proposal_factory.create(status=OrderProposal.Status.CONFIRMED)
        tp1 = order_proposal_factory.create(
            portfolio=tp0.portfolio,
            status=OrderProposal.Status.CONFIRMED,
            trade_date=(tp0.trade_date + BusinessMonthEnd(1)).date(),
        )
        tp2 = order_proposal_factory.create(
            portfolio=tp0.portfolio,
            status=OrderProposal.Status.CONFIRMED,
            trade_date=(tp1.trade_date + BusinessMonthEnd(1)).date(),
        )

        # Replay order proposals
        tp0.replay()

        # Expected calls to drift_weights
        expected_calls = [
            call(tp0.trade_date, tp1.trade_date - timedelta(days=1), stop_at_rebalancing=True),
            call(tp1.trade_date, tp2.trade_date - timedelta(days=1), stop_at_rebalancing=True),
            # call(tp2.trade_date, date.today(), stop_at_rebalancing=True),
        ]

        # Assert drift_weights was called as expected
        mock_fct.assert_has_calls(expected_calls)

        # Test stopping replay on a non-approved proposal
        tp1.status = OrderProposal.Status.FAILED
        tp1.save()
        expected_calls = [call(tp0.trade_date, tp1.trade_date - timedelta(days=1), stop_at_rebalancing=True)]
        mock_fct.assert_has_calls(expected_calls)

    # Test estimating shares for a trade
    @patch.object(OrderProposal, "get_portfolio_total_asset_value")
    def test_get_estimated_shares(
        self, mock_fct, order_proposal, order_factory, instrument_price_factory, instrument_factory
    ):
        """
        Verify shares estimation based on trade weighting and instrument price.
        """
        portfolio = order_proposal.portfolio
        instrument = instrument_factory.create(currency=portfolio.currency)
        underlying_quote_price = instrument_price_factory.create(instrument=instrument, date=order_proposal.trade_date)
        mock_fct.return_value = Decimal(1_000_000)  # 1 million cash
        trade = order_factory.create(
            order_proposal=order_proposal,
            value_date=order_proposal.trade_date,
            portfolio=portfolio,
            underlying_instrument=instrument,
        )
        trade.refresh_from_db()

        # Assert estimated shares are correctly calculated
        assert (
            order_proposal.get_estimated_shares(
                trade.weighting, trade.underlying_instrument, underlying_quote_price.net_value
            )
            == Decimal(1_000_000) * trade.weighting / underlying_quote_price.net_value
        )

    @patch.object(OrderProposal, "get_portfolio_total_asset_value")
    def test_get_estimated_target_cash(self, mock_fct, order_proposal, order_factory, cash_factory):
        order_proposal.portfolio.only_weighting = False
        order_proposal.portfolio.save()
        mock_fct.return_value = Decimal(1_000_000)  # 1 million cash
        cash = cash_factory.create(currency=order_proposal.portfolio.currency)
        order_factory.create(  # equity trade
            order_proposal=order_proposal,
            value_date=order_proposal.trade_date,
            portfolio=order_proposal.portfolio,
            weighting=Decimal("0.7"),
        )
        order_factory.create(  # cash trade
            order_proposal=order_proposal,
            value_date=order_proposal.trade_date,
            portfolio=order_proposal.portfolio,
            underlying_instrument=cash,
            weighting=Decimal("0.2"),
        )
        target_cash_position = order_proposal.get_estimated_target_cash()
        assert target_cash_position.weighting == Decimal("0.3")
        assert target_cash_position.initial_shares == Decimal(1_000_000) * Decimal("0.3")

    def test_order_proposal_update_inception_date(self, order_proposal_factory, portfolio, instrument_factory):
        # Check that if we create a prior order proposal, the instrument inception date is updated accordingly
        instrument = instrument_factory.create(inception_date=None)
        instrument.portfolios.add(portfolio)
        tp = order_proposal_factory.create(portfolio=portfolio)
        instrument.refresh_from_db()
        assert instrument.inception_date == tp.trade_date

        tp2 = order_proposal_factory.create(portfolio=portfolio, trade_date=(tp.trade_date - BDay(1)).date())
        instrument.refresh_from_db()
        assert instrument.inception_date == tp2.trade_date

    def test_get_round_lot_size(self, order_proposal, instrument):
        # without a round lot size, we expect no normalization of shares
        assert order_proposal.get_round_lot_size(Decimal("66"), instrument) == Decimal("66")
        instrument.round_lot_size = 100
        instrument.save()

        # if instrument has a round lot size different than 1, we expect different behavior based on whether shares is positive or negative
        assert order_proposal.get_round_lot_size(Decimal(66.0), instrument) == Decimal("100")
        assert order_proposal.get_round_lot_size(Decimal(-66.0), instrument) == Decimal(-66.0)
        assert order_proposal.get_round_lot_size(Decimal(-120), instrument) == Decimal(-200)

        # exchange can disable rounding based on the lot size
        instrument.exchange.apply_round_lot_size = False
        instrument.exchange.save()
        assert order_proposal.get_round_lot_size(Decimal("66"), instrument) == Decimal("66")

    @patch.object(OrderProposal, "get_portfolio_total_asset_value")
    def test_submit_round_lot_size(self, mock_fct, order_proposal, instrument_price_factory, order_factory):
        initial_shares = Decimal("70")
        price = instrument_price_factory.create(date=order_proposal.trade_date)
        net_value = round(price.net_value, 4)
        portfolio_value = initial_shares * net_value
        mock_fct.return_value = portfolio_value

        order_proposal.portfolio.only_weighting = False
        order_proposal.portfolio.save()
        instrument = price.instrument
        instrument.round_lot_size = 100
        instrument.save()
        trade = order_factory.create(
            shares=initial_shares,
            order_proposal=order_proposal,
            weighting=Decimal("1.0"),
            underlying_instrument=price.instrument,
            price=net_value,
        )
        warnings = order_proposal.submit()
        order_proposal.save()
        assert (
            len(warnings) == 1
        )  # ensure that submit returns a warning concerning the rounded trade based on the lot size
        trade.refresh_from_db()
        assert trade.shares == 100  # we expect the share to be transformed from 70 to 100 (lot size of 100)

    @patch.object(OrderProposal, "get_portfolio_total_asset_value")
    def test_submit_round_fractional_shares(
        self, mock_fct, instrument_price_factory, order_proposal, order_factory, asset_position_factory
    ):
        initial_shares = Decimal("5.6")
        price = instrument_price_factory.create(date=order_proposal.trade_date)
        net_value = round(price.net_value, 4)
        portfolio_value = initial_shares * net_value
        mock_fct.return_value = portfolio_value

        order_proposal.portfolio.only_weighting = False
        order_proposal.portfolio.save()

        trade = order_factory.create(
            shares=Decimal("5.6"),
            order_proposal=order_proposal,
            weighting=Decimal("1.0"),
            underlying_instrument=price.instrument,
            price=net_value,
        )
        order_proposal.submit()
        order_proposal.save()
        trade.refresh_from_db()
        assert trade.shares == 6  # we expect the fractional share to be rounded
        assert trade.weighting == round((trade.shares * net_value) / portfolio_value, 8)
        assert trade.weighting == round(
            Decimal("1") + ((Decimal("6") - initial_shares) * net_value) / portfolio_value, 8
        )  # we expect the weighting to be updated accrodingly

    def test_ex_post(
        self, instrument_factory, asset_position_factory, instrument_price_factory, order_proposal_factory, portfolio
    ):
        """
        Tests the ex-post rebalancing mechanism of a portfolio with two instruments.
        Verifies that weights are correctly recalculated after submitting and approving a order proposal.
        """

        # --- Create instruments ---
        msft = instrument_factory.create(currency=portfolio.currency)
        apple = instrument_factory.create(currency=portfolio.currency)

        # --- Key dates ---
        d1 = date(2025, 6, 24)
        d2 = date(2025, 6, 25)
        d3 = date(2025, 6, 26)
        d4 = date(2025, 6, 27)

        # --- Create MSFT prices ---
        msft_p1 = instrument_price_factory.create(instrument=msft, date=d1, net_value=Decimal("10"))
        msft_p2 = instrument_price_factory.create(instrument=msft, date=d2, net_value=Decimal("8"))
        msft_p3 = instrument_price_factory.create(instrument=msft, date=d3, net_value=Decimal("12"))
        msft_p4 = instrument_price_factory.create(instrument=msft, date=d4, net_value=Decimal("15"))  # noqa

        # Calculate MSFT returns between dates
        msft_r2 = msft_p2.net_value / msft_p1.net_value - Decimal("1")  # noqa
        msft_r3 = msft_p3.net_value / msft_p2.net_value - Decimal("1")

        # --- Create Apple prices (stable) ---
        apple_p1 = instrument_price_factory.create(instrument=apple, date=d1, net_value=Decimal("100"))
        apple_p2 = instrument_price_factory.create(instrument=apple, date=d2, net_value=Decimal("100"))
        apple_p3 = instrument_price_factory.create(instrument=apple, date=d3, net_value=Decimal("100"))
        apple_p4 = instrument_price_factory.create(instrument=apple, date=d4, net_value=Decimal("100"))  # noqa

        # Apple returns (always 0 since price is stable)
        apple_r2 = apple_p2.net_value / apple_p1.net_value - Decimal("1")  # noqa
        apple_r3 = apple_p3.net_value / apple_p2.net_value - Decimal("1")

        # --- Create positions on d2 ---
        msft_a2 = asset_position_factory.create(
            portfolio=portfolio,
            underlying_quote=msft,
            date=d2,
            initial_shares=10,
            weighting=Decimal("0.44"),
        )
        apple_a2 = asset_position_factory.create(
            portfolio=portfolio,
            underlying_quote=apple,
            date=d2,
            initial_shares=1,
            weighting=Decimal("0.56"),
        )

        # Check that initial weights sum to 1
        total_weight_d2 = msft_a2.weighting + apple_a2.weighting
        assert total_weight_d2 == pytest.approx(Decimal("1.0"), abs=Decimal("1e-6"))

        # --- Calculate total portfolio return between d2 and d3 ---
        portfolio_r3 = msft_a2.weighting * (Decimal("1.0") + msft_r3) + apple_a2.weighting * (
            Decimal("1.0") + apple_r3
        )

        # --- Create positions on d3 with weights adjusted for returns ---

        # Check that weights on d2 sum to 1
        total_weight_d2 = msft_a2.weighting + apple_a2.weighting
        assert total_weight_d2 == pytest.approx(Decimal("1.0"), abs=Decimal("1e-6"))

        # --- Create a order proposal on d3 ---
        order_proposal = order_proposal_factory.create(portfolio=portfolio, trade_date=d3)
        order_proposal.reset_orders()
        # Retrieve orders for each instrument
        orders = order_proposal.get_orders()
        trade_msft = orders.get(underlying_instrument=msft)
        trade_apple = orders.get(underlying_instrument=apple)
        # Check that trade weights are initially zero
        assert trade_msft.weighting == Decimal("0")
        assert trade_apple.weighting == Decimal("0")

        msft_drifted = msft_a2.weighting * (Decimal("1.0") + msft_r3) / portfolio_r3
        apple_drifted = apple_a2.weighting * (Decimal("1.0") + apple_r3) / portfolio_r3
        # --- Adjust trade weights to target 50% each ---
        target_weight = Decimal("0.5")
        trade_msft.weighting = target_weight - msft_drifted
        trade_msft.save()

        trade_apple.weighting = target_weight - apple_drifted
        trade_apple.save()
        orders = order_proposal.get_orders()
        trade_msft = orders.get(underlying_instrument=msft)
        trade_apple = orders.get(underlying_instrument=apple)

        # --- Check drift factors and effective weights ---
        assert trade_msft.daily_return == pytest.approx(msft_r3, abs=Decimal("1e-6"))
        assert trade_apple.daily_return == pytest.approx(apple_r3, abs=Decimal("1e-6"))

        assert trade_msft._effective_weight == pytest.approx(msft_drifted, abs=Decimal("1e-6"))
        assert trade_apple._effective_weight == pytest.approx(apple_drifted, abs=Decimal("1e-6"))

        # Check that the target weight is the sum of drifted weight and adjustment
        assert trade_msft._target_weight == pytest.approx(target_weight, abs=Decimal("1e-6"))
        assert trade_apple._target_weight == pytest.approx(target_weight, abs=Decimal("1e-6"))

        # --- Submit and approve the order proposal ---
        order_proposal.submit()
        order_proposal.save()
        order_proposal.approve()
        order_proposal.apply()
        order_proposal.save()

        # Final check that weights have been updated to 50%
        assert order_proposal.portfolio.assets.get(underlying_instrument=msft).weighting == pytest.approx(
            target_weight, abs=Decimal("1e-6")
        )
        assert order_proposal.portfolio.assets.get(underlying_instrument=apple).weighting == pytest.approx(
            target_weight, abs=Decimal("1e-6")
        )

    def test_replay_reset_draft_order_proposal(
        self, instrument_factory, instrument_price_factory, order_factory, order_proposal_factory
    ):
        instrument = instrument_factory.create()
        order_proposal = order_proposal_factory.create(trade_date=(date.today() - BDay(2)).date())
        instrument_price_factory.create(instrument=instrument, date=(date.today() - BDay(2)).date())
        instrument_price_factory.create(instrument=instrument, date=(date.today() - BDay(1)).date())
        instrument_price_factory.create(instrument=instrument, date=date.today())
        trade = order_factory.create(
            underlying_instrument=instrument,
            order_proposal=order_proposal,
            weighting=1,
        )
        order_proposal.submit()
        order_proposal.approve()
        order_proposal.confirm()
        order_proposal.save()

        draft_tp = order_proposal_factory.create(
            portfolio=order_proposal.portfolio, trade_date=(date.today() - BDay(1)).date()
        )
        assert not Order.objects.filter(order_proposal=draft_tp).exists()
        order_proposal.replay()

        assert Order.objects.filter(order_proposal=draft_tp).count() == 1
        assert Order.objects.get(
            order_proposal=draft_tp, underlying_instrument=trade.underlying_instrument
        ).weighting == Decimal("0")

    def test_order_submit_bellow_minimum_allowed_order_value(self, order_factory):
        order = order_factory.create(price=Decimal(1), weighting=Decimal(1), shares=Decimal(999))
        order.submit()
        order.save()
        assert order.shares == Decimal(999)
        assert order.weighting == Decimal(1)

        order.order_proposal.min_order_value = Decimal(1000)
        order.order_proposal.save()

        order.submit()
        order.save()
        assert order.shares == Decimal(0)
        assert order.weighting == Decimal(0)

    def test_order_submit_bellow_minimum_weighting(self, order_factory, order_proposal):
        o1 = order_factory.create(order_proposal=order_proposal, price=Decimal(1), weighting=Decimal("0.8"))
        o2 = order_factory.create(order_proposal=order_proposal, price=Decimal(1), weighting=Decimal("0.2"))
        order_proposal.submit()
        order_proposal.save()

        o1.refresh_from_db()
        o2.refresh_from_db()
        assert o1.weighting == Decimal("0.8")
        assert o2.weighting == Decimal("0.2")

        order_proposal.min_weighting = Decimal("0.21")
        order_proposal.backtodraft()
        order_proposal.submit()
        order_proposal.save()

        o1.refresh_from_db()
        o2.refresh_from_db()
        assert o1.weighting == Decimal("0.8")
        assert o2.weighting == Decimal("0")

        order_proposal.approve()
        order_proposal.apply()
        order_proposal.save()
        assert order_proposal.portfolio.assets.get(
            date=order_proposal.trade_date, underlying_quote=o1.underlying_instrument
        ).weighting == Decimal("0.8")
        assert order_proposal.portfolio.assets.get(
            date=order_proposal.trade_date, underlying_quote=order_proposal.cash_component
        ).weighting == Decimal("0.2")

    def test_reset_order_use_desired_target_weight(self, order_proposal, order_factory):
        order1 = order_factory.create(
            order_proposal=order_proposal, weighting=Decimal("0.5"), desired_target_weight=Decimal("0.7")
        )
        order2 = order_factory.create(
            order_proposal=order_proposal, weighting=Decimal("0.5"), desired_target_weight=Decimal("0.3")
        )
        order_proposal.submit()
        order_proposal.approve()
        order_proposal.apply()
        order_proposal.save()

        order1.refresh_from_db()
        order2.refresh_from_db()
        assert order1.desired_target_weight == Decimal("0.5")
        assert order2.desired_target_weight == Decimal("0.5")
        assert order1.weighting == Decimal("0.5")
        assert order2.weighting == Decimal("0.5")

        order1.desired_target_weight = Decimal("0.7")
        order2.desired_target_weight = Decimal("0.3")
        order1.save()
        order2.save()

        order_proposal.reset_orders(use_desired_target_weight=True)
        order1.refresh_from_db()
        order2.refresh_from_db()
        assert order1.weighting == Decimal("0.7")
        assert order2.weighting == Decimal("0.3")

    def test_reset_order_proposal_keeps_target_cash_weight(self, order_factory, order_proposal_factory):
        order_proposal = order_proposal_factory.create(
            total_cash_weight=Decimal("0.02")
        )  # create a OP with total cash weight of 2%

        # create orders that total weight account for only 50%
        order_factory.create(order_proposal=order_proposal, weighting=Decimal("0.3"))
        order_factory.create(order_proposal=order_proposal, weighting=Decimal("0.2"))

        order_proposal.reset_orders()
        assert order_proposal.get_orders().exclude(underlying_instrument__is_cash=True).aggregate(
            s=Sum("target_weight")
        )["s"] == Decimal("0.98"), "The total target weight leftover does not equal the stored total cash weight"

    def test_convert_to_portfolio_always_100percent(self, order_proposal, order_factory):
        o1 = order_factory.create(order_proposal=order_proposal, weighting=Decimal("0.5"))
        o2 = order_factory.create(order_proposal=order_proposal, weighting=Decimal("0.3"))
        portfolio = order_proposal.get_default_target_portfolio().build_cash(
            order_proposal.cash_component.id, order_proposal.total_effective_portfolio_weight
        )
        assert portfolio.positions_map[o1.underlying_instrument.id].weighting == Decimal("0.5")
        assert portfolio.positions_map[o2.underlying_instrument.id].weighting == Decimal("0.3")
        assert portfolio.positions_map[order_proposal.cash_component.id].weighting == Decimal("0.2")

    @patch.object(OrderProposal, "custodian_router", new_callable=PropertyMock)
    @patch.object(OrderProposal, "has_non_successful_checks", new_callable=PropertyMock)
    def test_can_execute(
        self, mock_has_non_successful_checks, mock_router, order_proposal, user_factory, mock_adapter
    ):
        user = user_factory.create()
        mock_router.return_value = mock_adapter
        mock_has_non_successful_checks.return_value = False
        order_proposal.status = OrderProposal.Status.APPROVED
        order_proposal.execution_status = ""

        assert order_proposal.can_execute(user) is True
        order_proposal.approver = user.profile
        assert order_proposal.can_execute(user) is False
        user.is_superuser = True
        assert order_proposal.can_execute(user) is True

        mock_router.return_value = None
        assert order_proposal.can_execute(user) is False

        mock_router.return_value = mock_adapter
        mock_has_non_successful_checks.return_value = True
        assert order_proposal.can_execute(user) is False

        mock_has_non_successful_checks.return_value = False
        order_proposal.status = OrderProposal.Status.PENDING
        assert order_proposal.can_execute(user) is False

        order_proposal.status = OrderProposal.Status.APPROVED
        order_proposal.execution_status = "something"
        assert order_proposal.can_execute(user) is False

    @patch.object(OrderProposal, "custodian_router", new_callable=PropertyMock)
    def test_refresh_execution_status(self, mock_custodian_router, order_proposal, mock_adapter):
        mock_custodian_router.return_value = mock_adapter
        mock_adapter.get_rebalance_status.return_value = (ExecutionStatus.PENDING, "detail")

        with patch.object(order_proposal, "save", wraps=order_proposal.save) as mock_save:
            order_proposal.refresh_execution_status()
            assert order_proposal.execution_status == ExecutionStatus.PENDING
            assert order_proposal.execution_status_detail == "detail"
            mock_save.assert_called_once()

    @patch.object(OrderProposal, "custodian_router", new_callable=PropertyMock)
    def test_cancel_rebalancing_success(self, mock_custodian_router, order_proposal, mock_adapter):
        mock_custodian_router.return_value = mock_adapter
        mock_adapter.cancel_rebalancing.return_value = True
        with patch.object(order_proposal, "save", wraps=order_proposal.save) as mock_save:
            result = order_proposal.cancel_rebalancing()
            assert result is True
            assert order_proposal.execution_status == ExecutionStatus.CANCELLED
            assert order_proposal.execution_comment == ""
            assert order_proposal.execution_status_detail == ""
            mock_save.assert_called_once()

    @patch.object(OrderProposal, "custodian_router", new_callable=PropertyMock)
    def test_cancel_rebalancing_failure(self, mock_custodian_router, order_proposal, mock_adapter):
        mock_custodian_router.return_value = mock_adapter
        mock_adapter.cancel_rebalancing.return_value = False
        with patch.object(order_proposal, "save", wraps=order_proposal.save) as mock_save:
            result = order_proposal.cancel_rebalancing()
            assert result is False
            # No change in status or calls to save
            mock_save.assert_not_called()

    @patch.object(OrderProposal, "custodian_router", new_callable=PropertyMock)
    @patch.object(OrderProposal, "prepare_orders_for_execution")
    @patch.object(OrderProposal, "handle_orders")
    def test_execute_orders_success(self, mock_handler_error, mock_fct, mock_router, order_proposal, mock_adapter):
        mock_router.return_value = mock_adapter
        # Arrange
        orders = ["order1", "order2"]
        mock_fct.return_value = orders
        confirmed_orders = ["confirmed1", "confirmed2"]
        status = ExecutionStatus.PENDING
        comment = "Success"
        mock_adapter.submit_rebalancing.return_value = (confirmed_orders, (status, comment))

        # Act
        with patch.object(order_proposal, "save") as mock_save:
            order_proposal.execute_orders(prioritize_target_weight=True)

            # Assert
            mock_fct.assert_called_once_with(prioritize_target_weight=True)
            mock_adapter.submit_rebalancing.assert_called_once_with(orders)
            mock_handler_error.assert_called_once_with(confirmed_orders)
            assert order_proposal.execution_status == status
            assert order_proposal.execution_comment == comment
            mock_save.assert_called_once()

    @patch.object(OrderProposal, "custodian_router", new_callable=PropertyMock)
    @patch.object(OrderProposal, "prepare_orders_for_execution")
    @patch.object(OrderProposal, "handle_orders")
    def test_execute_orders_on_failure(self, mock_handler_error, mock_fct, mock_router, order_proposal, mock_adapter):
        mock_router.return_value = mock_adapter
        # Arrange
        orders = ["order1", "order2"]
        mock_fct.return_value = orders
        mock_adapter.submit_rebalancing.side_effect = RoutingException("Failure!")

        # Act
        with patch.object(order_proposal, "save") as mock_save:
            order_proposal.execute_orders(prioritize_target_weight=True)

            # Assert
            mock_fct.assert_called_once_with(prioritize_target_weight=True)
            mock_adapter.submit_rebalancing.assert_called_once_with(orders)
            mock_handler_error.assert_not_called()
            assert order_proposal.execution_status == ExecutionStatus.FAILED
            assert order_proposal.execution_comment == "Failure!"
            mock_save.assert_called_once()

    def test_prepare_orders_for_execution(self, order_proposal, order_factory, instrument_factory, equity_factory):
        invalid_equity = equity_factory.create(refinitiv_identifier_code=None, ticker=None, sedol=None)
        exotic_instrument = instrument_factory.create()
        cash_instrument = instrument_factory.create(is_cash=True)
        order_valid = order_factory.create(
            order_proposal=order_proposal,
            weighting=Decimal(0.6),
            shares=Decimal(800),
            execution_instruction=ExecutionInstruction.LIMIT_ORDER,
            underlying_instrument=equity_factory.create(),
        )
        order_valid_but_unsupported_asset_class = order_factory.create(
            order_proposal=order_proposal,
            weighting=Decimal(0.6),
            shares=Decimal(800),
            execution_instruction=ExecutionInstruction.LIMIT_ORDER,
            underlying_instrument=exotic_instrument,
        )
        order_invalid_instrument = order_factory.create(
            order_proposal=order_proposal,
            weighting=Decimal(0.3),
            shares=Decimal(800),
            underlying_instrument=invalid_equity,
        )
        order_zero_delta = order_factory.create(
            order_proposal=order_proposal,
            weighting=Decimal(0),
            shares=Decimal(0),
            underlying_instrument=equity_factory.create(),
        )
        order_cash = order_factory.create(
            order_proposal=order_proposal,
            weighting=Decimal(0.1),
            shares=Decimal(200),
            underlying_instrument=cash_instrument,
        )

        orders_dto = order_proposal.prepare_orders_for_execution()
        assert len(orders_dto) == 1
        order = orders_dto[0]
        assert order.refinitiv_identifier_code == order_valid.underlying_instrument.refinitiv_identifier_code
        assert (
            order.bloomberg_ticker
            == order_valid.underlying_instrument.ticker
            + " "
            + order_valid.underlying_instrument.exchange.bbg_composite
        )
        assert order.sedol == order_valid.underlying_instrument.sedol
        assert order.execution_instruction == ExecutionInstruction.LIMIT_ORDER
        assert order.target_shares == order_valid.shares
        assert order.shares == order_valid.shares
        assert order.weighting == order_valid.weighting
        assert order.target_weight == order_valid.weighting
        assert order.trade_date == order_proposal.trade_date

        order_invalid_instrument.refresh_from_db()
        order_valid_but_unsupported_asset_class.refresh_from_db()
        order_zero_delta.refresh_from_db()
        order_cash.refresh_from_db()

        assert order_zero_delta.execution_status == Order.ExecutionStatus.IGNORED
        assert order_cash.execution_status == Order.ExecutionStatus.IGNORED
        assert order_invalid_instrument.execution_status == Order.ExecutionStatus.FAILED
        assert order_invalid_instrument.execution_comment == "Underlying instrument does not have a valid identifier."
        assert order_valid_but_unsupported_asset_class.execution_status == Order.ExecutionStatus.FAILED
        assert order_valid_but_unsupported_asset_class.execution_comment.startswith("Unsupported asset class")

    @patch.object(OrderProposal, "get_portfolio_total_asset_value")
    def test_handle_orders(self, mock_fct, order_proposal, order_factory):
        o1 = order_factory.create(
            order_proposal=order_proposal, weighting=Decimal("0.8"), shares=Decimal(800), price=Decimal(2)
        )
        o2 = order_factory.create(
            order_proposal=order_proposal, weighting=Decimal("0.2"), shares=Decimal(200), price=Decimal(2)
        )
        portfolio_value = Decimal(800) * Decimal(2) + Decimal(200) * Decimal(2)
        mock_fct.return_value = portfolio_value

        expected_shares = round(800 * 2 / 1.2)
        order_proposal.handle_orders(
            [
                OrderDTO(
                    id=o1.id,
                    asset_class=OrderDTO.AssetType.EQUITY,
                    target_weight=0.8,
                    weighting=0.8,
                    trade_date=o1.value_date,
                    execution_price=1.2,
                    shares=expected_shares,  # we simulate a market fluctuation
                    target_shares=expected_shares,
                    comment="some comment",
                )
            ]
        )

        o1.refresh_from_db()
        with pytest.raises(Order.DoesNotExist):
            o2.refresh_from_db()

        assert o1.execution_status == Order.ExecutionStatus.CONFIRMED
        assert o1.execution_comment == "some comment"

        # We do not update these fields anymore, we keep the test around in case it comes back
        # assert o1.price == Decimal("1.2")  # check the the new execution price was updated
        # assert (
        #     o1.shares == expected_shares
        # )  # check that the new shares based on the execution price got updated as well
        # assert (
        #     o1.weighting == round(Decimal(expected_shares * 1.2), 2) / portfolio_value
        # )  # weighting should change slightly as we round the number of shares

        assert order_proposal.orders.get(underlying_instrument__is_cash=True).weighting == Decimal("1") - o1.weighting



import logging
from datetime import datetime

from django.conf import settings
from requests import HTTPError

from wbportfolio.api_clients.ubs import UBSNeoAPIClient
from wbportfolio.pms.typing import Order

from .. import ExecutionInstruction, ExecutionStatus, RoutingException
from . import BaseCustodianAdapter

logger = logging.getLogger("oms")


ASSET_CLASS_MAP = {
    Order.AssetType.EQUITY: "EQUITY",
    Order.AssetType.AMERICAN_DEPOSITORY_RECEIPT: "EQUITY",
}  # API can support BOND, FUTURE, OPTION, and DYNAMIC_STRATEGY
ASSET_CLASS_MAP_INV = {
    v: k for k, v in ASSET_CLASS_MAP.items()
}  # API can support BOND, FUTURE, OPTION, and DYNAMIC_STRATEGY

STATUS_MAP = {
    "Amend Pending": ExecutionStatus.PENDING,
    "Cancel Pending": ExecutionStatus.PENDING,
    "Cancelled": ExecutionStatus.CANCELLED,
    "Complete": ExecutionStatus.COMPLETED,
    "Complete (Order Cancelled)": ExecutionStatus.COMPLETED,
    "Complete (Partial Fill)": ExecutionStatus.COMPLETED,
    "In Draft": ExecutionStatus.IN_DRAFT,
    "Pending Approval": ExecutionStatus.PENDING,
    "Pending Execution": ExecutionStatus.PENDING,
    "Rebalance Cancelled": ExecutionStatus.CANCELLED,
    "Rebalance Cancelled (Executing partially)": ExecutionStatus.CANCELLED,
    "Rejected": ExecutionStatus.REJECTED,
    "Rejection Acknowledged": ExecutionStatus.PENDING,
    "Waiting for Response": ExecutionStatus.PENDING,
}
EXECUTION_INSTRUCTION_MAP = {
    ExecutionInstruction.MARKET_ON_CLOSE: "MARKET_ON_CLOSE",
    ExecutionInstruction.GUARANTEED_MARKET_ON_CLOSE: "GUARANTEED_MARKET_ON_CLOSE",
    ExecutionInstruction.GUARANTEED_MARKET_ON_OPEN: "GUARANTEED_MARKET_ON_OPEN",
    ExecutionInstruction.GPW_MARKET_ON_CLOSE: "GPW_MARKET_ON_CLOSE",
    ExecutionInstruction.MARKET_ON_OPEN: "MARKET_ON_OPEN",
    ExecutionInstruction.IN_LINE_WITH_VOLUME: "IN_LINE_WITH_VOLUME",
    ExecutionInstruction.LIMIT_ORDER: "LIMIT_ORDER",
    ExecutionInstruction.VWAP: "VWAP",
    ExecutionInstruction.TWAP: "TWAP",
}
EXECUTION_INSTRUCTION_MAP_INV = {v: k for k, v in EXECUTION_INSTRUCTION_MAP.items()}


class CustodianAdapter(BaseCustodianAdapter):
    client: UBSNeoAPIClient

    def __init__(self, *args, raise_exception: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.raise_exception = raise_exception

    def _handle_response(self, res):
        logger.info(res["message"])
        if errors := res.get("errors"):
            logger.warning(errors)
            if self.raise_exception:
                raise RoutingException(errors)

    def _serialize_execution_instruction(
        self, execution_instruction: ExecutionInstruction, execution_parameters: dict
    ):
        repr = EXECUTION_INSTRUCTION_MAP[execution_instruction]
        if execution_parameters:
            if execution_instruction == ExecutionInstruction.IN_LINE_WITH_VOLUME:
                repr += f':{execution_parameters["percent"]:.0f%}'
            elif execution_instruction == ExecutionInstruction.LIMIT_ORDER:
                repr += f':{execution_parameters["price"]:.1f}'
                if good_for_date := execution_parameters.get("good_for_date"):
                    repr += f",{good_for_date}"
            elif (
                execution_instruction == ExecutionInstruction.VWAP
                or execution_instruction == ExecutionInstruction.TWAP
            ):
                repr += f':{execution_parameters["period"]},{execution_parameters["time"]}'
        return repr

    def serialize_orders(self, orders: list[Order]) -> list[dict[str, str]]:
        items = []
        for order in orders:
            if order.refinitiv_identifier_code:
                identifier_type, identifier = "RIC", order.refinitiv_identifier_code
            elif order.bloomberg_ticker:
                identifier_type, identifier = "BBTICKER", order.bloomberg_ticker
            else:
                identifier_type, identifier = "SEDOL", order.sedol
            item = {
                "assetClass": ASSET_CLASS_MAP[order.asset_class],
                "identifierType": identifier_type,
                "identifier": identifier,
                "executionInstruction": self._serialize_execution_instruction(
                    order.execution_instruction, order.execution_instruction_parameters
                ),
                "userElementId": str(order.id),
                "tradeDate": order.trade_date.strftime("%Y-%m-%d"),
            }
            if order.shares:
                item["sharesToTrade"] = str(order.shares)
            else:
                item["targetWeight"] = str(order.target_weight * 100)
            items.append(item)
        return items

    def deserialize_items(self, items: list[dict[str, str]]):
        orders = []
        for item in items:
            orders.append(
                Order(
                    id=item.get("userElementId"),
                    asset_class=ASSET_CLASS_MAP_INV[item.get("assetClass")],
                    refinitiv_identifier_code=item.get(
                        "ric", item["identifier"] if item.get("identifierType") == "RIC" else None
                    ),
                    bloomberg_ticker=item["identifier"] if item.get("identifierType") == "BBTICKER" else None,
                    sedol=item["identifier"] if item.get("identifierType") == "SEDOL" else None,
                    trade_date=datetime.strptime(item.get("tradeDate"), "%Y-%m-%d"),
                    target_weight=float(item["targetWeight"]) / 100 if "targetWeight" in item else None,
                    shares=float(item["sharesToTrade"]) if "sharesToTrade" in item else None,
                    execution_instruction=EXECUTION_INSTRUCTION_MAP_INV[item["executionInstruction"]],
                )
            )
        return orders

    def authenticate(self) -> bool:
        """
        Authenticate or renew tokens with the custodian API.
        Raises an exception if authentication fails.
        """
        self.client = UBSNeoAPIClient(settings.UBS_NEO_API_TOKEN)
        return True

    def is_valid(self) -> bool:
        """
        Check whether the given isin is valid and can be rebalanced
        """

        try:
            status_res = self.client.get_rebalance_service_status()

            isin_res = self.client.get_rebalance_status_for_isin(self.isin)
            self._handle_response(status_res)
            self._handle_response(isin_res)
            return (
                status_res["status"] == UBSNeoAPIClient.SUCCESS_VALUE
                and isin_res["status"] == UBSNeoAPIClient.SUCCESS_VALUE
            )
        except (HTTPError, KeyError) as e:
            logger.warning(f"Couldn't validate adapter: {str(e)}")
            return False

    def submit_rebalancing(
        self, items: list[dict[str, str]], as_draft: bool = True
    ) -> tuple[list[dict[str, str]], str]:
        """
        Submit a rebalance order for the certificate.
        """
        if not as_draft:
            res = self.client.submit_rebalance(self.isin, items)
        else:
            res = self.client.save_draft(self.isin, items)
        self._handle_response(res)
        return res["rebalanceItems"], res["message"]

    def cancel_current_rebalancing(self) -> bool:
        """
        Cancel an existing rebalance order identified by ISIN.
        """
        try:
            res = self.client.cancel_rebalance(self.isin)
            self._handle_response(res)
            return res["status"] == UBSNeoAPIClient.SUCCESS_VALUE
        except (HTTPError, KeyError):
            return False

    def get_rebalance_status(self) -> tuple[ExecutionStatus, str]:
        res = self.client.get_rebalance_status_for_isin(self.isin)
        self._handle_response(res)
        status = res.get("rebalanceStatus", "")
        return STATUS_MAP.get(status, ExecutionStatus.UNKNOWN), status

    def get_current_rebalancing(self) -> list[dict[str, str]]:
        """
        Fetch the current rebalance request details for a certificate.
        """
        res = self.client.get_current_rebalance_request(self.isin)
        self._handle_response(res)
        return res["rebalanceItems"]



import logging
import math
from contextlib import suppress
from datetime import date, timedelta
from decimal import Decimal
from typing import Any, Self, TypeVar

import pandas as pd
from celery import shared_task
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import DatabaseError, models
from django.db.models import (
    F,
    OuterRef,
    Q,
    QuerySet,
    Subquery,
    Sum,
    Value,
)
from django.db.models.functions import Coalesce, Round
from django.db.models.signals import post_save, pre_delete
from django.dispatch import receiver
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy
from django_fsm import FSMField, transition
from pandas._libs.tslibs.offsets import BDay
from requests import HTTPError
from wbcompliance.models.risk_management.mixins import RiskCheckMixin
from wbcore.contrib.authentication.models import User
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton
from wbcore.models import WBModel
from wbcore.signals import pre_merge
from wbcore.utils.models import CloneMixin
from wbcore.workers import Queue
from wbfdm.enums import MarketData
from wbfdm.models import InstrumentPrice
from wbfdm.models.instruments.instruments import Cash, Instrument
from wbfdm.signals import investable_universe_updated

from wbportfolio.models.asset import AssetPosition
from wbportfolio.models.roles import PortfolioRole
from wbportfolio.pms.analytics.portfolio import Portfolio as AnalyticPortfolio
from wbportfolio.pms.typing import Order as OrderDTO
from wbportfolio.pms.typing import Portfolio as PortfolioDTO

from ...order_routing import ExecutionStatus, RoutingException
from ...order_routing.router import Router
from .. import Portfolio
from .orders import Order

logger = logging.getLogger("pms")

SelfOrderProposal = TypeVar("SelfOrderProposal", bound="OrderProposal")


class OrderProposal(CloneMixin, RiskCheckMixin, WBModel):
    trade_date = models.DateField(verbose_name="Trading Date")

    class Status(models.TextChoices):
        DRAFT = "DRAFT", "Draft"
        PENDING = "PENDING", "Pending"
        APPROVED = "APPROVED", "Approved"
        DENIED = "DENIED", "Denied"
        EXECUTION = "EXECUTION", "Execution"
        CONFIRMED = "CONFIRMED", "Confirmed"
        FAILED = "FAILED", "Failed"

    comment = models.TextField(default="", verbose_name="Order Comment", blank=True)
    status = FSMField(default=Status.DRAFT, choices=Status.choices, verbose_name="Status")
    rebalancing_model = models.ForeignKey(
        "wbportfolio.RebalancingModel",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="order_proposals",
        verbose_name="Rebalancing Model",
        help_text="Rebalancing Model that generates the target portfolio",
    )
    portfolio = models.ForeignKey(
        "wbportfolio.Portfolio", related_name="order_proposals", on_delete=models.PROTECT, verbose_name="Portfolio"
    )
    creator = models.ForeignKey(
        "directory.Person",
        blank=True,
        null=True,
        related_name="order_proposals",
        on_delete=models.PROTECT,
        verbose_name="Owner",
    )
    approver = models.ForeignKey(
        "directory.Person",
        blank=True,
        null=True,
        related_name="approver_order_proposals",
        on_delete=models.PROTECT,
        verbose_name="Approver",
    )
    min_order_value = models.IntegerField(
        default=0, verbose_name="Minimum Order Value", help_text="Minimum Order Value in the Portfolio currency"
    )
    min_weighting = models.DecimalField(
        max_digits=9,
        decimal_places=Order.ORDER_WEIGHTING_PRECISION,
        default=Decimal(0),
        help_text="The minimum weight allowed for this order proposal ",
        verbose_name="Minimum Weight",
        validators=[MinValueValidator(Decimal("0")), MaxValueValidator(Decimal("1"))],
    )

    total_cash_weight = models.DecimalField(
        default=Decimal("0"),
        decimal_places=4,
        max_digits=5,
        verbose_name="Total Cash Weight",
        help_text="The desired percentage for the cash component. The remaining percentage (100% minus this value) will be allocated to total target weighting. Default is 0%.",
        validators=[MinValueValidator(Decimal("0")), MaxValueValidator(Decimal("1"))],
    )
    total_effective_portfolio_contribution = models.DecimalField(
        default=Decimal("1"),
        max_digits=Order.ORDER_WEIGHTING_PRECISION * 2 + 3,
        decimal_places=Order.ORDER_WEIGHTING_PRECISION * 2,
    )
    execution_status = models.CharField(
        blank=True, default="", choices=ExecutionStatus.choices, verbose_name="Execution Status"
    )
    execution_status_detail = models.CharField(blank=True, default="", verbose_name="Execution Status Detail")
    execution_comment = models.CharField(blank=True, default="", verbose_name="Execution Comment")

    class Meta:
        verbose_name = "Order Proposal"
        verbose_name_plural = "Order Proposals"
        constraints = [
            models.UniqueConstraint(
                fields=["portfolio", "trade_date"],
                name="unique_order_proposal",
            ),
        ]

        notification_types = [
            create_notification_type(
                "wbportfolio.order_proposal.push_model_changes",
                "Push Model Changes",
                "Sends a notification when a the change/orders are pushed to modeled after portfolios",
                True,
                True,
                True,
            )
        ]

    def __str__(self) -> str:
        return f"{self.portfolio.name}: {self.trade_date} ({self.status})"

    def save(self, *args, **kwargs):
        # if the order proposal is created, we default these fields with the portfolio default value for automatic value assignement
        if not self.id and not self.min_order_value:
            self.min_order_value = self.portfolio.default_order_proposal_min_order_value
        if not self.id and not self.min_weighting:
            self.min_weighting = self.portfolio.default_order_proposal_min_weighting
        if not self.id and not self.total_cash_weight:
            self.total_cash_weight = self.portfolio.default_order_proposal_total_cash_weight
        # if a order proposal is created before the existing earliest order proposal, we automatically shift the linked instruments inception date to allow automatic NAV computation since the new inception date
        if not self.portfolio.order_proposals.filter(trade_date__lt=self.trade_date).exists():
            # we need to set the inception date as the first order proposal trade date (and thus, the first position date). We expect a NAV at 100 then
            self.portfolio.instruments.filter(inception_date__gt=self.trade_date).update(
                inception_date=self.trade_date
            )
        super().save(*args, **kwargs)

    @property
    def check_evaluation_date(self):
        return self.trade_date

    @property
    def checked_object(self) -> Any:
        return self.portfolio

    @cached_property
    def portfolio_total_asset_value(self) -> Decimal:
        return self.get_portfolio_total_asset_value()

    @cached_property
    def last_effective_date(self) -> date:
        try:
            return self.portfolio.assets.filter(date__lt=self.trade_date).latest("date").date
        except AssetPosition.DoesNotExist:
            return self.value_date

    @property
    def custodian_router(self) -> Router | None:
        try:
            return Router(self.portfolio.get_authenticated_custodian_adapter(raise_exception=True))
        except ValueError as e:
            logger.warning("Error while instantiating custodian adapter: %s", e)

    @cached_property
    def value_date(self) -> date:
        return (self.trade_date - BDay(1)).date()

    @property
    def previous_order_proposal(self) -> SelfOrderProposal | None:
        future_proposals = OrderProposal.objects.filter(portfolio=self.portfolio).filter(
            trade_date__lt=self.trade_date, status=OrderProposal.Status.CONFIRMED
        )
        if future_proposals.exists():
            return future_proposals.latest("trade_date")
        return None

    @property
    def next_order_proposal(self) -> SelfOrderProposal | None:
        future_proposals = OrderProposal.objects.filter(portfolio=self.portfolio).filter(
            trade_date__gt=self.trade_date, status=OrderProposal.Status.CONFIRMED
        )
        if future_proposals.exists():
            return future_proposals.earliest("trade_date")
        return None

    @property
    def cash_component(self) -> Cash:
        return self.portfolio.cash_component

    @property
    def total_effective_portfolio_weight(self) -> Decimal:
        return Decimal("1.0")

    @property
    def total_expected_target_weight(self) -> Decimal:
        return self.total_effective_portfolio_weight - self.total_cash_weight

    @property
    def can_be_confirmed(self) -> bool:
        return self.portfolio.can_be_rebalanced and self.status == self.Status.APPROVED

    @property
    def can_be_applied(self):
        return not self.has_non_successful_checks and self.portfolio.is_manageable

    @cached_property
    def total_effective_portfolio_cash_weight(self) -> Decimal:
        return self.portfolio.assets.filter(
            models.Q(date=self.last_effective_date)
            & (models.Q(underlying_quote__is_cash=True) | models.Q(underlying_quote__is_cash_equivalent=True))
        ).aggregate(Sum("weighting"))["weighting__sum"] or Decimal("0")

    def get_portfolio_total_asset_value(self):
        return self.portfolio.get_total_asset_value(self.last_effective_date)
        # return self.orders.annotate(
        #     effective_shares=Coalesce(
        #         Subquery(
        #             AssetPosition.objects.filter(
        #                 underlying_quote=OuterRef("underlying_instrument"),
        #                 date=self.last_effective_date,
        #                 portfolio=self.portfolio,
        #             )
        #             .values("portfolio")
        #             .annotate(s=Sum("shares"))
        #             .values("s")[:1]
        #         ),
        #         Decimal(0),
        #     ),
        #     effective_total_value_fx_portfolio=F("effective_shares") * F("currency_fx_rate") * F("price"),
        # ).aggregate(s=Sum("effective_total_value_fx_portfolio"))["s"] or Decimal(0.0)

    def get_orders(self):
        # TODO Issue here: the cash is subqueried on the portfolio, on portfolio such as the fund, there is multiple cash component, that we exclude in the orders (and use a unique cash position instead)
        # so the subquery returns the previous position (probably USD), but is missing the other cash aggregation. We need to find a way to handle that properly

        orders = self.orders.all().annotate(
            total_effective_portfolio_contribution=Value(self.total_effective_portfolio_contribution),
            last_effective_date=Subquery(
                AssetPosition.unannotated_objects.filter(
                    date__lt=OuterRef("value_date"),
                    portfolio=OuterRef("portfolio"),
                )
                .order_by("-date")
                .values("date")[:1]
            ),
            previous_weight=models.Case(
                models.When(
                    underlying_instrument__is_cash=False,
                    then=Coalesce(
                        Subquery(
                            AssetPosition.unannotated_objects.filter(
                                underlying_quote=OuterRef("underlying_instrument"),
                                date=OuterRef("last_effective_date"),
                                portfolio=OuterRef("portfolio"),
                            )
                            .values("portfolio")
                            .annotate(s=Sum("weighting"))
                            .values("s")[:1]
                        ),
                        Decimal(0),
                    ),
                ),
                default=Value(self.total_effective_portfolio_cash_weight),
            ),
            contribution=F("previous_weight") * (F("daily_return") + Value(Decimal("1"))),
            effective_weight=Round(
                models.Case(
                    models.When(total_effective_portfolio_contribution=Value(Decimal("0")), then=Value(Decimal("0"))),
                    default=F("contribution") / F("total_effective_portfolio_contribution") - F("quantization_error"),
                ),
                precision=Order.ORDER_WEIGHTING_PRECISION,
            ),
            target_weight=Round(F("effective_weight") + F("weighting"), precision=Order.ORDER_WEIGHTING_PRECISION),
            effective_shares=Coalesce(
                Subquery(
                    AssetPosition.objects.filter(
                        underlying_quote=OuterRef("underlying_instrument"),
                        date=OuterRef("last_effective_date"),
                        portfolio=OuterRef("portfolio"),
                    )
                    .values("portfolio")
                    .annotate(s=Sum("shares"))
                    .values("s")[:1]
                ),
                Decimal(0),
            ),
            target_shares=F("effective_shares") + F("shares"),
        )

        return orders.annotate(
            has_warnings=models.Case(
                models.When(
                    (models.Q(price=0) & ~models.Q(target_weight=0)) | models.Q(target_weight__lt=0), then=Value(True)
                ),
                default=Value(False),
            ),
        )

    def get_trades_batch(self):
        return self._get_default_effective_portfolio().get_orders(self.get_target_portfolio())

    @property
    def can_be_executed(self) -> bool:
        return (
            self.custodian_router is not None
            and not self.has_non_successful_checks
            and self.status == self.Status.APPROVED
            and (not self.execution_status or self.execution_status == ExecutionStatus.CANCELLED)
        )

    def can_execute(self, user: User) -> bool:
        return (not self.approver or user.is_superuser or user != self.approver.user_account) and self.can_be_executed

    def prepare_orders_for_execution(self, prioritize_target_weight: bool = False) -> list[OrderDTO]:
        """Prepares executable orders by filtering and converting them for submission.

        Filters out cash instruments and orders with zero weighting and shares, then
        creates OrderDTOs for those having valid instrument identifiers. Orders with
        unsupported asset classes or missing identifiers are marked as ignored with comments.
        Updates ignored orders in bulk.

        Args:
            prioritize_target_weight: If True, prioritize target weight over share quantities
              when preparing order quantities.

        Returns:
            A list of OrderDTO objects ready for execution submission.
        """
        executable_orders = []
        updated_orders = []
        self.orders.update(execution_status=Order.ExecutionStatus.IGNORED)
        for order in (
            self.get_orders()
            .exclude(models.Q(underlying_instrument__is_cash=True) | (models.Q(weighting=0) & models.Q(shares=0)))
            .select_related("underlying_instrument")
        ):
            instrument = order.underlying_instrument
            asset_class = instrument.get_security_ancestor().instrument_type.key.upper()

            try:
                if instrument.refinitiv_identifier_code or instrument.ticker or instrument.sedol:
                    quantity = {"target_weight": float(order.target_weight)}
                    if not prioritize_target_weight and order.shares:
                        quantity["shares"] = float(order.shares)
                        quantity["target_shares"] = (
                            float(order.target_shares) if order.target_shares is not None else None
                        )

                    executable_orders.append(
                        OrderDTO(
                            id=order.id,
                            asset_class=OrderDTO.AssetType[asset_class],
                            weighting=float(order.weighting),
                            trade_date=order.value_date,
                            refinitiv_identifier_code=instrument.refinitiv_identifier_code,
                            bloomberg_ticker=instrument.bloomberg_ticker,
                            sedol=instrument.sedol,
                            execution_instruction=order.execution_instruction,
                            execution_instruction_parameters=order.execution_instruction_parameters,
                            **quantity,
                        )
                    )
                else:
                    order.execution_status = Order.ExecutionStatus.FAILED
                    order.execution_comment = "Underlying instrument does not have a valid identifier."
                    updated_orders.append(order)
            except (AttributeError, KeyError):
                order.execution_status = Order.ExecutionStatus.FAILED
                order.execution_comment = f"Unsupported asset class {asset_class.title()}."
                updated_orders.append(order)

        Order.objects.bulk_update(updated_orders, ["execution_status", "execution_comment"])
        return executable_orders

    def handle_orders(self, orders: list[OrderDTO]):
        """Updates order statuses based on confirmed execution results.

        For each confirmed order, updates the corresponding database record with its
        execution status, comment, and price when available. Orders not present in the
        confirmation list are marked as failed.

        Args:
            orders: List of confirmed order DTOs returned from the custodian router.
        """
        leftover_orders = self.orders.filter(underlying_instrument__is_cash=False).all()
        # portfolio_value = self.portfolio_total_asset_value

        for confirmed_order in orders:
            with suppress(Order.DoesNotExist):
                order = leftover_orders.get(id=confirmed_order.id)
                order.execution_status = Order.ExecutionStatus.CONFIRMED
                order.execution_comment = confirmed_order.comment
                # if execution_price := confirmed_order.execution_price:
                #     order.price = round(Decimal(execution_price), 2)
                #     order.execution_status = Order.ExecutionStatus.EXECUTED
                # if shares := confirmed_order.shares:
                #     order.set_shares(Decimal(shares), portfolio_value)
                # elif weighting := confirmed_order.weighting:
                #     order.set_weighting(Decimal(weighting), portfolio_value)
                order.save()
                leftover_orders = leftover_orders.exclude(id=order.id)

        leftover_orders.delete()
        self.refresh_cash_position()

    def execute_orders(self, prioritize_target_weight: bool = False):
        """Submits prepared orders for execution via the custodian router and updates status.

        Prepares orders based on the target weight priority, submits them for execution,
        handles confirmed orders on success, and records execution status and comments.
        Logs and marks the execution as failed if submission raises an error.

        Args:
            prioritize_target_weight: Whether to prioritize target weights when preparing orders.
        """
        self.status = self.Status.EXECUTION
        orders = self.prepare_orders_for_execution(prioritize_target_weight=prioritize_target_weight)
        try:
            confirmed_orders, (status, rebalancing_comment) = self.custodian_router.submit_rebalancing(orders)
            self.handle_orders(confirmed_orders)
        except (ValueError, RoutingException, HTTPError) as e:
            logger.error(
                "Order Execution: routing or network exception ecountered.",
                extra={"order_proposal": self, "detail": e},
            )
            status = ExecutionStatus.FAILED
            rebalancing_comment = str(e)
        self.execution_status = status
        self.execution_comment = rebalancing_comment
        self.save()

    def refresh_execution_status(self):
        """Updates execution status from the custodian router and saves the model.

        Retrieves the latest rebalance status and details, assigns them to the instance,
        and persists changes to the database.
        """
        self.execution_status, self.execution_status_detail = self.custodian_router.get_rebalance_status()
        if self.execution_status == ExecutionStatus.COMPLETED:
            self.execution_status = self.Status.CONFIRMED
        self.save()

    def cancel_rebalancing(self):
        """Cancels the ongoing rebalance via the custodian router and updates the model.

        If cancellation succeeds, clears execution details, marks the status as cancelled,
        saves the instance, and returns the cancellation status.
        """
        cancel_rebalancing_status = self.custodian_router.cancel_rebalancing()
        if cancel_rebalancing_status:
            (
                self.execution_comment,
                self.execution_status_detail,
                self.execution_status,
            ) = (
                "",
                "",
                ExecutionStatus.CANCELLED,
            )
            self.save()
        return cancel_rebalancing_status

    def get_target_portfolio(self):
        positions = []
        instrument_ids = []
        for order in self.get_orders():
            pos = order.to_dto()
            instrument_ids.append(pos.underlying_instrument)
            positions.append(pos)

        # insert latest market data
        df = pd.DataFrame(
            Instrument.objects.filter(id__in=instrument_ids).dl.market_data(
                [MarketData.MARKET_CAPITALIZATION_CONSOLIDATED, MarketData.VOLUME, MarketData.CLOSE],
                from_date=self.trade_date - timedelta(days=50),
                to_date=self.trade_date,
                target_currency="USD",
            )
        )
        df["volume_50d"] = df["volume"]
        df = (
            df[
                [
                    "valuation_date",
                    "instrument_id",
                    "volume",
                    "volume_50d",
                    "close",
                    "market_capitalization_consolidated",
                ]
            ]
            .sort_values(by="valuation_date")
            .groupby("instrument_id")
            .agg(
                {
                    "volume": "last",
                    "volume_50d": "mean",
                    "close": "last",
                    "market_capitalization_consolidated": "last",
                }
            )
        )
        df["volume_usd"] = df.volume * df.close

        for pos in positions:
            if pos.underlying_instrument in df.index:
                pos.market_capitalization_usd = df.loc[pos.underlying_instrument, "market_capitalization_consolidated"]
                pos.volume_usd = (
                    df.loc[pos.underlying_instrument, "volume"] * df.loc[pos.underlying_instrument, "close"]
                )
                if pos.shares:
                    if volume_50d := df.loc[pos.underlying_instrument, "volume_50d"]:
                        pos.daily_liquidity = float(pos.shares) / volume_50d / 0.33
                    if pos.market_capitalization_usd:
                        pos.market_share = (
                            float(pos.shares)
                            * df.loc[pos.underlying_instrument, "close"]
                            / pos.market_capitalization_usd
                        )

        return PortfolioDTO(positions)

    # Start tools methods
    def _clone(self, **kwargs) -> SelfOrderProposal:
        """
        Method to clone self as a new order proposal. It will automatically shift the order date if a proposal already exists
        Args:
            **kwargs: The keyword arguments
        Returns:
            The cloned order proposal
        """
        trade_date = kwargs.get("clone_date", self.trade_date)

        # Find the next valid order date
        while OrderProposal.objects.filter(portfolio=self.portfolio, trade_date=trade_date).exists():
            trade_date += timedelta(days=1)

        order_proposal_clone = OrderProposal.objects.create(
            trade_date=trade_date,
            comment=kwargs.get("clone_comment", self.comment),
            status=OrderProposal.Status.DRAFT,
            rebalancing_model=self.rebalancing_model,
            portfolio=self.portfolio,
            creator=self.creator,
        )
        for order in self.orders.all():
            order.id = None
            order.order_proposal = order_proposal_clone
            order.save()

        return order_proposal_clone

    def normalize_orders(self, total_cash_weight: Decimal):
        """
        Normalize the orders to accomodate the given cash weight
        """
        self.total_cash_weight = total_cash_weight
        self.reset_orders()

    def fix_quantization(self):
        if self.orders.exists():
            orders = self.get_orders()
            t_weight = orders.aggregate(models.Sum("effective_weight"))["effective_weight__sum"] or Decimal("0.0")
            quantization_error = orders.aggregate(models.Sum("quantization_error"))[
                "quantization_error__sum"
            ] or Decimal("0.0")
            # we handle quantization error due to the decimal max digits. In that case, we take the biggest order (highest weight) and we remove the quantization error
            if t_weight and (
                quantize_error := ((t_weight + quantization_error) - self.total_effective_portfolio_weight)
            ):
                biggest_order = orders.exclude(underlying_instrument__is_cash=True).latest("effective_weight")
                biggest_order.quantization_error = quantize_error
                biggest_order.save()

    def get_default_target_portfolio(self, use_desired_target_weight: bool = False, **kwargs) -> PortfolioDTO:
        if self.rebalancing_model:
            params = {}
            if rebalancer := getattr(self.portfolio, "automatic_rebalancer", None):
                params.update(rebalancer.parameters)
            params.update(kwargs)
            return self.rebalancing_model.get_target_portfolio(
                self.portfolio, self.trade_date, self.value_date, **params
            )
        portfolio = self._get_default_effective_portfolio()
        for pos in portfolio.positions:
            with suppress(Order.DoesNotExist):
                order = self.orders.get(underlying_instrument_id=pos.underlying_instrument)
                if use_desired_target_weight:
                    pos.weighting = order.desired_target_weight
                else:
                    pos.weighting += order.weighting
        return portfolio

    def _get_default_previous_portfolio(self):
        """
        Converts the internal portfolio state and pending orders into a PortfolioDTO.

        Returns:
            PortfolioDTO: Object that encapsulates all portfolio positions.
        """
        portfolio = {}

        # 1. Gather all non-cash, positively weighted assets from the existing portfolio.
        for asset in self.portfolio.assets.filter(
            date=self.last_effective_date,
            weighting__gt=0,
        ):
            portfolio[asset.underlying_quote] = {
                "shares": asset._shares,
                "weighting": asset.weighting,
                "price": asset._price,
                "currency_fx_rate": asset._currency_fx_rate,
            }

        # 2. Add or update non-cash orders, possibly overriding weights.
        for order in self.get_orders().filter(
            underlying_instrument__is_cash=False, underlying_instrument__is_cash_equivalent=False
        ):
            portfolio[order.underlying_instrument] = {
                "weighting": order._previous_weight,
                "shares": order._effective_shares,
                "price": order.price,
                "currency_fx_rate": order.currency_fx_rate,
            }
        positions = []

        # 5. Build PositionDTO objects for all instruments.
        for instrument, row in portfolio.items():
            # Assemble the position object
            pos = Order.create_dto(
                instrument,
                row["weighting"],
                row["price"],
                self.last_effective_date,
                shares=row["shares"],
                currency_fx_rate=row["currency_fx_rate"],
            )
            positions.append(pos)
        return PortfolioDTO(positions)

    def _get_default_effective_portfolio(self):
        portfolio = self._get_default_previous_portfolio()
        self._set_returns(portfolio)
        return portfolio.build_cash(self.cash_component.id, self.total_effective_portfolio_weight)

    def _set_returns(self, portfolio: PortfolioDTO):
        try:
            self.portfolio.builder.load_returns(
                portfolio.positions_map.keys(), self.last_effective_date, self.trade_date, use_dl=True
            )
            analytic_portfolio = AnalyticPortfolio(
                X=self.portfolio.builder.returns.loc[: self.trade_date, :],
                weights={i: float(pos.weighting) for i, pos in portfolio.positions_map.items()},
            )
            last_returns, contribution = analytic_portfolio.get_contributions()

            last_returns = last_returns.to_dict()
            effective_weights = analytic_portfolio.get_next_weights()
        except IndexError:
            effective_weights, last_returns, contribution = {}, {}, 1
        self.total_effective_portfolio_contribution = Decimal(contribution)
        for pos in portfolio.positions:
            pos.daily_return = Decimal(last_returns.get(pos.underlying_instrument, 0))
            pos.weighting = Decimal(effective_weights.get(pos.underlying_instrument, pos.weighting))

    def reset_orders(
        self,
        effective_portfolio: PortfolioDTO
        | None = None,  # we need to have this parameter as sometime we want to get the effective portfolio from drifted weight (unsaved)
        target_portfolio: PortfolioDTO | None = None,
        use_desired_target_weight: bool = False,
    ):
        """
        Will delete all existing orders and recreate them from the method `create_or_update_trades`
        """
        if self.rebalancing_model:
            self.orders.all().delete()
        else:
            self.orders.filter(underlying_instrument__is_cash=True).delete()
        # delete all existing orders
        # Get effective and target portfolio
        if not effective_portfolio:
            effective_portfolio = self._get_default_previous_portfolio()
        self._set_returns(effective_portfolio)
        effective_portfolio = effective_portfolio.build_cash(
            self.cash_component.id, self.total_effective_portfolio_weight
        )
        if not target_portfolio:
            target_portfolio = self.get_default_target_portfolio(use_desired_target_weight=use_desired_target_weight)
        if self.total_cash_weight:
            target_portfolio = target_portfolio.normalize_cash(self.total_cash_weight)
        target_portfolio = target_portfolio.build_cash(self.cash_component.id, self.total_effective_portfolio_weight)
        # ensure we have all prices before loading orders
        self.portfolio.builder.load_returns(
            target_portfolio.positions_map.keys(), self.last_effective_date, self.trade_date, use_dl=True
        )
        if target_portfolio:
            objs = []
            portfolio_value = self.portfolio_total_asset_value
            self.orders.update(quantization_error=0)
            for order_dto in effective_portfolio.get_orders(target_portfolio):
                instrument = Instrument.objects.get(id=order_dto.underlying_instrument)
                # we cannot do a bulk-create because Order is a multi table inheritance
                weighting = round(order_dto.delta_weight, Order.ORDER_WEIGHTING_PRECISION)
                daily_return = order_dto.daily_return
                try:
                    order = self.orders.get(underlying_instrument=instrument)
                    order.daily_return = daily_return
                except Order.DoesNotExist:
                    order = Order(
                        underlying_instrument=instrument,
                        order_proposal=self,
                        value_date=self.trade_date,
                        weighting=weighting,
                        daily_return=daily_return,
                    )
                order.order_type = Order.get_type(
                    weighting, round(order_dto.effective_weight, 8), round(order_dto.target_weight, 8)
                )
                order.quantization_error = order_dto.effective_quantization_error
                if order_dto.price:
                    order.price = order_dto.price
                order.pre_save(load_returns=False)
                order.set_weighting(weighting, portfolio_value)
                order.desired_target_weight = order_dto.target_weight

                # if we cannot automatically find a price, we consider the stock is invalid and we sell it
                if not order.price and order.weighting > 0:
                    order.price = Decimal("0.0")
                    order.weighting = -order_dto.effective_weight
                objs.append(order)
            Order.objects.bulk_create(
                objs,
                update_fields=[
                    "value_date",
                    "weighting",
                    "daily_return",
                    "currency_fx_rate",
                    "order_type",
                    "portfolio",
                    "price",
                    "price_gross",
                    "desired_target_weight",
                    "quantization_error",
                    "shares",
                ],
                unique_fields=["order_proposal", "underlying_instrument"],
                update_conflicts=True,
                batch_size=1000,
            )
        # final sanity check to make sure invalid order with effective and target weight of 0 are automatically removed:
        self.get_orders().exclude(underlying_instrument__is_cash=True).filter(
            target_weight=0, effective_weight=0
        ).delete()
        self.get_orders().filter(target_weight=0).exclude(effective_shares=0).update(shares=-F("effective_shares"))
        self.fix_quantization()
        self.save()

    def refresh_cash_position(self):
        self.total_cash_weight = self.total_effective_portfolio_weight - (
            self.get_orders().filter(underlying_instrument__is_cash=False).aggregate(s=Sum("target_weight"))["s"]
            or Decimal("0")
        )
        cash_order = None
        try:
            cash_order = Order.objects.get(order_proposal=self, underlying_instrument=self.cash_component)
        except Order.DoesNotExist:
            if self.total_cash_weight:
                cash_order = Order.objects.create(
                    order_proposal=self, underlying_instrument=self.cash_component, weighting=Decimal("0")
                )
        if cash_order:
            cash_order.weighting = self.total_cash_weight - cash_order._previous_weight
            cash_order.save()

    def refresh_returns(self):
        weights = {
            row[0]: float(row[1]) for row in self.get_orders().values_list("underlying_instrument", "previous_weight")
        }
        last_returns, contribution = self.portfolio.get_analytic_portfolio(
            self.value_date, weights=weights, use_dl=True
        ).get_contributions()
        last_returns = last_returns.to_dict()
        orders_to_update = []
        for order in self.orders.all():
            with suppress(KeyError):
                order.price = self.portfolio.builder.prices[self.trade_date][order.underlying_instrument.id]
            try:
                order.daily_return = last_returns[order.underlying_instrument.id]
            except KeyError:
                order.daily_return = Decimal("1.0")
            order.quantization_error = Decimal("0")
            orders_to_update.append(order)
        Order.objects.bulk_update(orders_to_update, ["daily_return", "price", "quantization_error"])
        self.total_effective_portfolio_contribution = Decimal(contribution)
        self.save()
        # ensure that sell orders keep having target weight at zero (might happens when returns are refreshed expost)
        for order in self.get_orders().filter(Q(order_type=Order.Type.SELL) & ~Q(weighting=-F("target_weight"))):
            order.weighting = -order.effective_weight
            order.save()

        # At this point, user needs to manually modify the orders in order to account for ex-post change. I am not sure we should we quantization at that point. To be monitored
        # self.fix_quantization()

    def replay(
        self,
        broadcast_changes_at_date: bool = True,
        reapply_order_proposal: bool = False,
        synchronous: bool = False,
        **reset_order_kwargs,
    ):
        last_order_proposal = self
        last_order_proposal_created = False
        max_date = (date.today() - BDay(1)).date()
        self.portfolio.load_builder_returns((self.trade_date - BDay(3)).date(), max_date)
        while last_order_proposal and last_order_proposal.status == OrderProposal.Status.CONFIRMED:
            last_order_proposal.portfolio = self.portfolio  # we set the same ptf reference
            if not last_order_proposal_created:
                if reapply_order_proposal or last_order_proposal.rebalancing_model:
                    logger.info(f"Replaying order proposal {last_order_proposal}")
                    last_order_proposal.apply_workflow(
                        silent_exception=True, force_reset_order=True, **reset_order_kwargs
                    )
                    last_order_proposal.save()
                else:
                    logger.info(f"Resetting order proposal {last_order_proposal}")
                    last_order_proposal.reset_orders(**reset_order_kwargs)
                if last_order_proposal.status != OrderProposal.Status.CONFIRMED:
                    break
            next_order_proposal = last_order_proposal.next_order_proposal
            if next_order_proposal:
                next_trade_date = next_order_proposal.trade_date - timedelta(days=1)
            elif next_expected_rebalancing_date := self.portfolio.get_next_rebalancing_date(
                last_order_proposal.trade_date
            ):
                next_trade_date = (
                    next_expected_rebalancing_date + timedelta(days=7)
                )  # we don't know yet if rebalancing is valid and can be executed on `next_expected_rebalancing_date`, so we add safety window of 7 days
            else:
                next_trade_date = max_date
            next_trade_date = min(next_trade_date, max_date)
            gen = self.portfolio.drift_weights(
                last_order_proposal.trade_date, next_trade_date, stop_at_rebalancing=True
            )
            try:
                while True:
                    self.portfolio.builder.add(next(gen))
            except StopIteration as e:
                overriding_order_proposal = e.value

            self.portfolio.builder.bulk_create_positions(
                delete_leftovers=True,
            )
            for draft_tp in OrderProposal.objects.filter(
                portfolio=self.portfolio,
                trade_date__gt=last_order_proposal.trade_date,
                trade_date__lte=next_trade_date,
                status=OrderProposal.Status.DRAFT,
            ):
                draft_tp.reset_orders()
            if overriding_order_proposal:
                last_order_proposal_created = True
                last_order_proposal = overriding_order_proposal
            else:
                last_order_proposal_created = False
                last_order_proposal = next_order_proposal
        self.portfolio.builder.schedule_change_at_dates(
            synchronous=synchronous, broadcast_changes_at_date=broadcast_changes_at_date
        )

    def invalidate_future_order_proposal(self):
        # Delete all future automatic order proposals and set the manual one into a draft state
        self.portfolio.order_proposals.filter(
            trade_date__gt=self.trade_date, rebalancing_model__isnull=False, comment="Automatic rebalancing"
        ).delete()
        for future_order_proposal in self.portfolio.order_proposals.filter(
            trade_date__gt=self.trade_date, status=OrderProposal.Status.CONFIRMED
        ):
            future_order_proposal.revert()
            future_order_proposal.save()

    def get_estimated_shares(
        self, weight: Decimal, underlying_quote: Instrument, quote_price: Decimal
    ) -> Decimal | None:
        """
        Estimates the number of shares for a order based on the given weight and underlying quote.

        This method calculates the estimated shares by dividing the order's total value in the portfolio's currency by the price of the underlying quote in the same currency. It handles currency conversion and suppresses any ValueError that might occur during the price retrieval.

        Args:
            weight (Decimal): The weight of the order.
            underlying_quote (Instrument): The underlying instrument for the order.

        Returns:
            Decimal | None: The estimated number of shares or None if the calculation fails.
        """
        # Retrieve the price of the underlying quote on the order date TODO: this is very slow and probably due to the to_date argument to the dl which slowdown drastically the query

        # if an order exists for this estimation and the target weight is 0, then we return the inverse of the effective shares
        with suppress(Order.DoesNotExist):
            order = self.get_orders().get(underlying_instrument=underlying_quote)
            if order.target_weight == 0:
                return -order.effective_shares
        # Calculate the order's total value in the portfolio's currency
        trade_total_value_fx_portfolio = self.portfolio_total_asset_value * weight

        # Convert the quote price to the portfolio's currency
        price_fx_portfolio = quote_price * underlying_quote.currency.convert(
            self.trade_date, self.portfolio.currency, exact_lookup=False
        )
        # If the price is valid, calculate and return the estimated shares
        if price_fx_portfolio:
            return trade_total_value_fx_portfolio / price_fx_portfolio

    def get_round_lot_size(self, shares: Decimal, underlying_quote: Instrument) -> Decimal:
        if (round_lot_size := underlying_quote.round_lot_size) != 1 and (
            not underlying_quote.exchange or underlying_quote.exchange.apply_round_lot_size
        ):
            if shares > 0:
                shares = math.ceil(shares / round_lot_size) * round_lot_size
            elif abs(shares) > round_lot_size:
                shares = math.floor(shares / round_lot_size) * round_lot_size
        return shares

    def get_estimated_target_cash(self, target_cash_weight: Decimal | None = None) -> AssetPosition:
        """
        Estimates the target cash weight and shares for a order proposal.

        This method calculates the target cash weight by summing the weights of cash orders and adding any leftover weight from non-cash orders. It then estimates the target shares for this cash component if the portfolio is not only weighting-based.

        Args:
            target_cash_weight (Decimal): the expected target cash weight (Optional). If not provided, we estimate from the existing orders

        Returns:
            tuple[Decimal, Decimal]: A tuple containing the target cash weight and the estimated target shares.
        """
        # Retrieve orders with base information
        orders = self.get_orders()
        # Calculate the total target weight of all orders
        total_target_weight = orders.filter(
            underlying_instrument__is_cash=False, underlying_instrument__is_cash_equivalent=False
        ).aggregate(s=models.Sum("target_weight"))["s"] or Decimal(0)
        if target_cash_weight is None:
            target_cash_weight = Decimal("1") - total_target_weight

        # Initialize target shares to zero
        total_target_shares = Decimal(0)

        # Get or create a cash component for the portfolio's currency
        cash_component = self.cash_component
        # If the portfolio is not only weighting-based, estimate the target shares for the cash component
        if not self.portfolio.only_weighting:
            # Estimate the target shares for the cash component
            with suppress(ValueError):
                total_target_shares = self.get_estimated_shares(target_cash_weight, cash_component, Decimal("1.0"))

        # otherwise, we create a new position
        underlying_quote_price = InstrumentPrice.objects.get_or_create(
            instrument=cash_component,
            date=self.trade_date,
            calculated=False,
            defaults={"net_value": Decimal(1)},
        )[0]
        return AssetPosition(
            underlying_quote=cash_component,
            portfolio_created=None,
            portfolio=self.portfolio,
            date=self.trade_date,
            weighting=target_cash_weight,
            initial_price=underlying_quote_price.net_value,
            initial_shares=total_target_shares,
            asset_valuation_date=self.trade_date,
            underlying_quote_price=underlying_quote_price,
            currency=cash_component.currency,
            is_estimated=False,
        )

    # WORKFLOW METHODS
    @transition(
        field=status,
        source=Status.DRAFT,
        target=Status.PENDING,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:order",),
                icon=WBIcon.SEND.icon,
                key="submit",
                label="Submit",
                action_label="Submit",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def submit(self, by=None, description=None, pretrade_check: bool = True, **kwargs):
        orders = []
        orders_validation_warnings = []
        qs = self.get_orders()
        for order in qs:
            order_warnings = order.submit(
                by=by, description=description, portfolio_total_asset_value=self.portfolio_total_asset_value, **kwargs
            )

            if order_warnings:
                orders_validation_warnings.extend(order_warnings)
            orders.append(order)

        Order.objects.bulk_update(orders, ["shares", "weighting", "desired_target_weight"])
        if pretrade_check:
            self.evaluate_pretrade_checks()
        else:
            self.refresh_cash_position()
        return orders_validation_warnings

    def can_submit(self):
        errors = dict()
        return errors

    @transition(
        field=status,
        source=Status.PENDING,
        target=Status.APPROVED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and not instance.has_non_successful_checks,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:order",),
                icon=WBIcon.APPROVE.icon,
                key="approve",
                label="Approve",
                action_label="Approve",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def approve(self, by=None, replay: bool = True, **kwargs):
        if by:
            self.approver = getattr(by, "profile", None)
        elif not self.approver:
            self.approver = self.creator
        if self.portfolio.can_be_rebalanced:
            self.apply()
            if replay:
                replay_as_task.apply_async(
                    (self.id,),
                    {
                        "user_id": by.id if by else None,
                        "broadcast_changes_at_date": False,
                        "reapply_order_proposal": True,
                    },
                    countdown=10,
                )
        if by and self.custodian_router:
            for user in User.objects.exclude(id=by.id).filter(
                profile__in=PortfolioRole.portfolio_managers(), is_active=True
            ):
                send_notification(
                    code="wbportfolio.portfolio.action_done",
                    title="An Order Proposal was approved and is waiting execution",
                    body=f"The order proposal {self} has been approved by {by.profile.full_name} and is now pending execution. Please review the orders carefully and proceed with execution if appropriate.",
                    user=user,
                    reverse_name="wbportfolio:orderproposal-detail",
                    reverse_args=[self.id],
                )

    @transition(
        field=status,
        source=Status.APPROVED,
        target=Status.CONFIRMED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and instance.portfolio.can_be_rebalanced,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:order",),
                icon=WBIcon.LOCK.icon,
                key="confirm",
                label="Confirm",
                action_label="Lock order proposal",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def confirm(self, by=None, replay: bool = True, **kwargs):
        self.refresh_cash_position()
        if self.portfolio.can_be_rebalanced:
            self.apply()
            if replay:
                replay_as_task.apply_async(
                    (self.id,),
                    {
                        "user_id": by.id if by else None,
                        "broadcast_changes_at_date": False,
                        "reapply_order_proposal": True,
                    },
                    countdown=10,
                )

    @transition(
        field=status,
        source=Status.PENDING,
        target=Status.DENIED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and not instance.has_non_successful_checks,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:order",),
                icon=WBIcon.DENY.icon,
                key="deny",
                label="Deny",
                action_label="Deny",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def deny(self, by=None, description=None, **kwargs):
        pass

    def can_deny(self):
        pass

    def apply(self):
        # We validate order which will create or update the initial asset positions
        if not self.portfolio.can_be_rebalanced:
            raise ValueError("Non-Rebalanceable portfolio cannot be traded manually.")
        self.portfolio.load_builder_returns(self.trade_date, self.trade_date)
        # We do not want to create the estimated cash position if there is not orders in the order proposal (shouldn't be possible anyway)
        target_portfolio = self.get_target_portfolio()
        assets = {i: float(pos.weighting) for i, pos in target_portfolio.positions_map.items()}
        self.portfolio.builder.add((self.trade_date, assets)).bulk_create_positions(
            force_save=True, is_estimated=False, delete_leftovers=True
        )

    @transition(
        field=status,
        source=Status.PENDING,
        target=Status.DRAFT,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        )
        and instance.has_all_check_completed
        or not instance.checks.exists(),  # we wait for all checks to succeed before proposing the back to draft transition
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:order",),
                icon=WBIcon.UNDO.icon,
                key="backtodraft",
                label="Back to Draft",
                action_label="backtodraft",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def backtodraft(self, **kwargs):
        self.checks.delete()

    def can_backtodraft(self):
        pass

    @transition(
        field=status,
        source=Status.CONFIRMED,
        target=Status.DRAFT,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:order",),
                icon=WBIcon.REGENERATE.icon,
                key="revert",
                label="Revert",
                action_label="revert",
                description_fields="<p>Unapply orders and move everything back to draft (i.e. The underlying asset positions will change like the orders were never applied)</p>",
            )
        },
    )
    def revert(self, **kwargs):
        self.approver = None
        self.portfolio.assets.filter(date=self.trade_date, is_estimated=False).update(
            is_estimated=True
        )  # we delete the existing portfolio as it has been reverted

    def can_revert(self):
        errors = dict()
        if not self.portfolio.can_be_rebalanced:
            errors["portfolio"] = [
                gettext_lazy(
                    "The portfolio needs to be a model portfolio in order to revert this order proposal manually"
                )
            ]
        return errors

    def apply_workflow(
        self,
        apply_automatically: bool = True,
        silent_exception: bool = False,
        force_reset_order: bool = False,
        **reset_order_kwargs,
    ):
        # before, we need to save all positions in the builder first because effective weight depends on it
        self.portfolio.load_builder_returns(self.trade_date, self.trade_date)
        self.portfolio.builder.bulk_create_positions(delete_leftovers=True)
        if self.status == OrderProposal.Status.CONFIRMED:
            logger.info("Reverting order proposal ...")
            self.revert()
        if self.status == OrderProposal.Status.DRAFT:
            if (
                self.rebalancing_model or force_reset_order
            ):  # if there is no position (for any reason) or we the order proposal has a rebalancer model attached (orders are computed based on an aglo), we reapply this order proposal
                logger.info("Resetting orders ...")
                try:  # we silent any validation error while setting proposal, because if this happens, we assume the current order proposal state if valid and we continue to batch compute
                    self.reset_orders(**reset_order_kwargs)
                except (ValidationError, DatabaseError) as e:
                    self.status = OrderProposal.Status.FAILED
                    if not silent_exception:
                        raise ValidationError(e) from e
                    return
            logger.info("Submitting order proposal ...")
            self.submit(pretrade_check=False)
        if apply_automatically:
            logger.info("Applying order proposal ...")
            if self.status == OrderProposal.Status.PENDING:
                self.approve(replay=False)
            else:
                self.apply()
            self.status = self.Status.CONFIRMED

    # End FSM logics

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbportfolio:orderproposal"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbportfolio:orderproposalrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{_portfolio.name}} ({{trade_date}})"

    @classmethod
    def build(
        cls,
        trade_date: date,
        portfolio,
        target_portfolio: PortfolioDTO,
        creator: User | None = None,
        approve_automatically: bool = True,
    ) -> Self:
        order_proposal, _ = OrderProposal.objects.update_or_create(
            portfolio=portfolio,
            trade_date=trade_date,
            defaults={"status": OrderProposal.Status.DRAFT, "creator": creator.profile if creator else None},
        )
        order_proposal.reset_orders(target_portfolio=target_portfolio)
        if approve_automatically:
            order_proposal.submit()
            order_proposal.approve(by=creator)
            if portfolio.can_be_rebalanced:
                order_proposal.apply()
            order_proposal.save()
        return order_proposal

    def push_to_dependant_portfolios(
        self, only_portfolios: QuerySet[Portfolio] | None = None, **build_kwargs
    ) -> list[Self]:
        order_proposals = []
        for rel in self.portfolio.get_model_portfolio_relationships(self.trade_date):
            existing_order_proposal = OrderProposal.objects.filter(
                portfolio=rel.portfolio, trade_date=self.trade_date
            ).first()
            # we allow push only on existing draft order proposal
            dependency_portfolio = rel.dependency_portfolio
            if (
                (only_portfolios is None or rel.portfolio in only_portfolios)
                and (not existing_order_proposal or existing_order_proposal.status == OrderProposal.Status.DRAFT)
                and dependency_portfolio.assets.filter(date=self.trade_date).exists()
            ):
                target_portfolio = dependency_portfolio._build_dto(self.trade_date)
                order_proposals.append(
                    OrderProposal.build(self.trade_date, rel.portfolio, target_portfolio, **build_kwargs)
                )
        return order_proposals

    def evaluate_pretrade_checks(self, asynchronously: bool = True):
        self.checks.all().delete()
        self.refresh_cash_position()
        self.evaluate_active_rules(
            self.trade_date,
            self.get_target_portfolio(),
            asynchronously=asynchronously,
            ignore_breached_objects=list(
                Instrument.objects.filter(id__in=self.orders.filter(weighting=0).values("underlying_instrument"))
            ),
        )


@receiver(post_save, sender="wbportfolio.OrderProposal")
def post_fail_order_proposal(sender, instance: OrderProposal, created, raw, **kwargs):
    # if we have a order proposal in a fail state, we ensure that all future existing order proposal are either deleted (automatic one) or set back to draft
    if not raw and instance.status == OrderProposal.Status.FAILED:
        # we delete all order proposal that have a rebalancing model and are marked as "automatic" (quite hardcoded yet)
        instance.invalidate_future_order_proposal()


@shared_task(queue=Queue.DEFAULT.value)
def replay_as_task(order_proposal_id, user_id: int | None = None, **kwargs):
    order_proposal = OrderProposal.objects.get(id=order_proposal_id)
    order_proposal.replay(**kwargs)
    if user_id:
        body = f'We’ve successfully replayed your order proposal for "{order_proposal.portfolio}" from {order_proposal.trade_date:%Y-%m-%d}. You can now review its updated composition.'
        user = User.objects.get(id=user_id)
        if order_proposal.portfolio.builder.excluded_positions:
            excluded_quotes = []
            for batch in order_proposal.portfolio.builder.excluded_positions.values():
                for pos in batch:
                    excluded_quotes.append(pos.underlying_instrument)
            body += "<p><strong>Note</strong></p><p>While replaying and drifting the portfolio, we excluded the positions from the following quotes because of missing price</p> <ul>"
            for excluded_quote in set(excluded_quotes):
                body += f"<li>{excluded_quote}</li>"
            body += "</ul>"
            order_proposal.portfolio.builder.clear()
        send_notification(
            code="wbportfolio.portfolio.action_done",
            title="Order Proposal Replay Completed",
            body=body,
            user=user,
            reverse_name="wbportfolio:portfolio-detail",
            reverse_args=[order_proposal.portfolio.id],
        )


@shared_task(queue=Queue.HIGH_PRIORITY.value)
def execute_orders_as_task(order_proposal_id: int, prioritize_target_weight: bool = False, **kwargs):
    order_proposal = OrderProposal.objects.get(id=order_proposal_id)
    order_proposal.execute_orders()


@shared_task(queue=Queue.BACKGROUND.value)
def push_model_change_as_task(
    model_order_proposal_id: int,
    user_id: int | None = None,
    only_for_portfolio_ids: list[int] | None = None,
    approve_automatically: bool = False,
):
    # not happy with that but we will keep it for the MVP lifecycle
    model_order_proposal = OrderProposal.objects.get(id=model_order_proposal_id)
    user = User.objects.get(id=user_id) if user_id else None
    params = dict(approve_automatically=approve_automatically, creator=user)
    only_portfolios = None
    if only_for_portfolio_ids:
        only_portfolios = Portfolio.objects.filter(id__in=only_for_portfolio_ids)

    order_proposals = model_order_proposal.push_to_dependant_portfolios(only_portfolios=only_portfolios, **params)
    product_html_list = "<ul>\n"
    for order_proposal in order_proposals:
        product_html_list += f"<li>{order_proposal.portfolio}</li>\n"

    product_html_list += "</ul>"
    if user:
        send_notification(
            code="wbportfolio.order_proposal.push_model_changes",
            title="Portfolio Model changes are pushed to dependant portfolios",
            body=f"""
    <p>The latest updates to the portfolio model <strong>{model_order_proposal.portfolio}</strong> have been successfully applied to the associated portfolios, and corresponding orders have been created.</p>
    <p>To proceed with executing these orders, please review the following related portfolios: </p>
            {product_html_list}
            """,
            user=user,
        )


@receiver(investable_universe_updated, sender="wbfdm.Instrument")
def update_exante_order_proposal_returns(*args, end_date: date | None = None, **kwargs):
    for op in OrderProposal.objects.filter(trade_date__gte=end_date):
        op.refresh_returns()


@receiver(pre_delete, sender=OrderProposal)
def post_delete_adjustment(sender, instance: OrderProposal, **kwargs):
    for check in instance.checks.all():
        check.delete()


@receiver(pre_merge, sender="directory.Person")
def pre_merge_person_orderproposal(sender: models.Model, merged_object, main_object, **kwargs):
    OrderProposal.objects.filter(creator=merged_object).update(creator=main_object)
    OrderProposal.objects.filter(approver=merged_object).update(approver=main_object)



# Generated by Django 5.2.9 on 2025-12-15 11:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0094_alter_portfolio_depends_on_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='portfolio',
            name='only_keep_essential_positions',
            field=models.BooleanField(default=False, help_text='Only keep essential positions when drifting positions'),
        ),
    ]



from rest_framework import serializers
from wbcore import serializers as wb_serializers
from wbcore.contrib.directory.serializers import PersonRepresentationSerializer
from wbfdm.serializers import InvestableInstrumentRepresentationSerializer

from wbportfolio.models import PortfolioRole


class PortfolioRoleModelSerializer(wb_serializers.ModelSerializer):
    """Model Serializer for Portfolio Roles"""

    _person = PersonRepresentationSerializer(source="person")
    _instrument = InvestableInstrumentRepresentationSerializer(source="instrument")

    def create(self, validated_data):
        instrument = validated_data.get("instrument", None)
        role_type = validated_data.get("role_type")

        if instrument is not None and role_type in [
            PortfolioRole.RoleType.MANAGER,
            PortfolioRole.RoleType.RISK_MANAGER,
        ]:
            raise serializers.ValidationError(
                {"instrument": [PortfolioRole.default_error_messages["manager"].format(model="instrument")]}
            )

        return super().create(validated_data)

    def update(self, instance, validated_data):
        instrument = validated_data.get("instrument", None)
        role_type = validated_data.get("role_type", instance.role_type)

        if instrument is not None and role_type in [
            PortfolioRole.RoleType.MANAGER,
            PortfolioRole.RoleType.RISK_MANAGER,
        ]:
            raise serializers.ValidationError(
                {"instrument": [PortfolioRole.default_error_messages["manager"].format(model="instrument")]}
            )

        return super().update(instance, validated_data)

    class Meta:
        model = PortfolioRole
        fields = (
            "id",
            "role_type",
            "person",
            "_person",
            "start",
            "end",
            "weighting",
            "instrument",
            "_instrument",
        )


class PortfolioRoleProjectModelSerializer(PortfolioRoleModelSerializer):
    """Model Serializer for Portfolio Roles for Products"""

    role_type = wb_serializers.ChoiceField(choices=PortfolioRole.RoleType.choices)



from contextlib import suppress

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.dispatch import receiver
from wbfdm.models import Instrument, InstrumentType

from wbportfolio.models import InstrumentPortfolioThroughModel, Product, ProductGroup


def _compute_str(instrument):
    with suppress(ObjectDoesNotExist):
        if instrument.instrument_type == InstrumentType.PRODUCT_GROUP:
            return ProductGroup.objects.get(id=instrument.id).compute_str()
        elif instrument.instrument_type == InstrumentType.PRODUCT:
            return Product.objects.get(id=instrument.id).compute_str()
    return instrument.get_compute_str()


@receiver(models.signals.class_prepared)
def add_to_instrument(sender, **kwargs):
    Instrument.add_to_class("portfolio", property(InstrumentPortfolioThroughModel.get_portfolio))
    Instrument.add_to_class("primary_portfolio", property(InstrumentPortfolioThroughModel.get_primary_portfolio))
    Instrument.add_to_class("compute_str", _compute_str)



from wbcore import viewsets
from wbcore.contrib.permission.permissions import InternalUserPermissionMixin

from wbportfolio.models import (
    InstrumentPortfolioThroughModel,
    PortfolioInstrumentPreferredClassificationThroughModel,
)
from wbportfolio.serializers import (
    InstrumentPortfolioThroughModelSerializer,
    InstrumentPreferedClassificationThroughProductModelSerializer,
)

from .configs import (
    InstrumentPortfolioThroughPortfolioModelDisplayConfig,
    InstrumentPortfolioThroughPortfolioModelEndpointConfig,
    PortfolioInstrumentPreferredClassificationThroughDisplayConfig,
    PortfolioInstrumentPreferredClassificationThroughEndpointConfig,
)


class InstrumentPreferedClassificationThroughProductModelViewSet(InternalUserPermissionMixin, viewsets.ModelViewSet):
    serializer_class = InstrumentPreferedClassificationThroughProductModelSerializer
    queryset = PortfolioInstrumentPreferredClassificationThroughModel.objects.all()

    search_fields = ("instrument__computed_str", "classification__computed_str")
    ordering_fields = ordering = ["instrument__computed_str"]
    filterset_fields = {"instrument": ["exact"], "classification": ["exact"]}
    display_config_class = PortfolioInstrumentPreferredClassificationThroughDisplayConfig
    endpoint_config_class = PortfolioInstrumentPreferredClassificationThroughEndpointConfig

    def get_queryset(self):
        return super().get_queryset().filter(portfolio=self.kwargs["portfolio_id"])


class InstrumentPortfolioThroughPortfolioModelViewSet(InternalUserPermissionMixin, viewsets.ModelViewSet):
    serializer_class = InstrumentPortfolioThroughModelSerializer
    queryset = InstrumentPortfolioThroughModel.objects.all()

    search_fields = ("instrument__name",)
    ordering_fields = ["instrument__name"]
    ordering = ["instrument__name"]

    display_config_class = InstrumentPortfolioThroughPortfolioModelDisplayConfig
    endpoint_config_class = InstrumentPortfolioThroughPortfolioModelEndpointConfig

    def get_queryset(self):
        return super().get_queryset().filter(portfolio=self.kwargs["portfolio_id"])



import factory
from pandas.tseries.offsets import BDay
from wbfdm.factories.instrument_prices import InstrumentPriceFactory

from wbportfolio.models.transactions.claim import Claim


class ClaimFactory(factory.django.DjangoModelFactory):
    status = Claim.Status.DRAFT
    account = factory.SubFactory("wbcrm.factories.AccountFactory")
    trade = factory.SubFactory("wbportfolio.factories.CustomerTradeFactory")
    product = factory.LazyAttribute(lambda x: x.trade.product)
    claimant = factory.SubFactory("wbcore.contrib.directory.factories.entries.PersonFactory")
    creator = factory.SubFactory("wbcore.contrib.directory.factories.entries.PersonFactory")
    date = factory.LazyAttribute(lambda x: (x.trade.transaction_date - BDay(0)).date())
    bank = factory.Faker("company")
    reference = factory.Faker("company")
    shares = factory.LazyAttribute(lambda x: x.trade.shares)
    nominal_amount = factory.LazyAttribute(lambda x: x.shares * x.product.share_price)
    external_id = factory.Sequence(lambda n: f"{n:06}")

    @factory.post_generation
    def create_product_val(self, create, extracted, **kwargs):
        if extracted is None or extracted is True:
            InstrumentPriceFactory.create(instrument=self.product, date=self.date, calculated=False)

    class Meta:
        model = Claim
        skip_postgeneration_save = True


class PositiveClaimFactory(ClaimFactory):
    shares = factory.Faker("pydecimal", min_value=0, max_value=1000000)


class NegativeClaimFactory(ClaimFactory):
    shares = factory.Faker("pydecimal", min_value=-1000000, max_value=0)


class ApprovedClaimFactory(ClaimFactory):
    status = Claim.Status.APPROVED



from django.contrib import admin

from wbportfolio.models import Trade


@admin.register(Trade)
class TradeAdmin(admin.ModelAdmin):
    search_fields = ["portfolio__name", "underlying_instrument__computed_str", "bank"]
    list_filter = ("portfolio", "pending")
    list_display = (
        "transaction_subtype",
        "transaction_date",
        "underlying_instrument",
        "portfolio",
        "shares",
        "price",
        "total_value",
        "pending",
        "marked_for_deletion",
        "exclude_from_history",
        "import_source",
    )

    readonly_fields = [
        "total_value",
        "total_value_gross",
        "total_value_fx_portfolio",
        "total_value_gross_fx_portfolio",
        "created",
        "updated",
    ]
    fieldsets = (
        (
            "Transaction Information",
            {
                "fields": (
                    ("transaction_subtype",),
                    ("pending", "marked_for_deletion", "exclude_from_history"),
                    (
                        "portfolio",
                        "underlying_instrument",
                        "import_source",
                    ),
                    ("transaction_date", "book_date", "value_date"),
                    ("price", "currency", "currency_fx_rate"),
                    ("shares", "total_value", "total_value_gross"),
                    ("total_value_fx_portfolio", "total_value_gross_fx_portfolio"),
                    ("register", "custodian", "bank", "external_id", "external_id_alternative"),
                    ("created", "updated"),
                    ("comment",),
                )
            },
        ),
    )
    autocomplete_fields = ["portfolio", "underlying_instrument", "currency", "register", "custodian"]
    ordering = ("-transaction_date",)
    raw_id_fields = ["import_source", "underlying_instrument", "portfolio", "register", "custodian"]



# Generated by Django 5.0.12 on 2025-11-06 12:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0090_dividendtransaction_price_fx_portfolio_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='order',
            name='execution_confirmed',
        ),
        migrations.AddField(
            model_name='order',
            name='execution_instruction',
            field=models.CharField(choices=[('MARKET_ON_CLOSE', 'Market On Close'), ('GUARANTEED_MARKET_ON_CLOSE', 'Guaranteed Market On Close'), ('GUARANTEED_MARKET_ON_OPEN', 'Guaranteed Market On Open'), ('GPW_MARKET_ON_CLOSE', 'GPW Market On Close'), ('MARKET_ON_OPEN', 'Market On Open'), ('IN_LINE_WITH_VOLUME', 'In Line With Volume'), ('LIMIT_ORDER', 'Limit Order'), ('VWAP', 'VWAP'), ('TWAP', 'TWAP')], default='MARKET_ON_CLOSE', max_length=26, verbose_name='Execution Instruction'),
        ),
        migrations.AddField(
            model_name='order',
            name='execution_instruction_parameters',
            field=models.JSONField(blank=True, default=dict, verbose_name='Execution Instruction Parameters'),
        ),
        migrations.AddField(
            model_name='order',
            name='execution_status',
            field=models.CharField(choices=[('PENDING', 'Pending'), ('CONFIRMED', 'Confirmed'), ('EXECUTED', 'Executed'), ('FAILED', 'Failed'), ('IGNORED', 'Ignored')], default='PENDING', max_length=12, verbose_name='Execution Status'),
        ),
    ]



from typing import TYPE_CHECKING, Any

from rest_framework.reverse import reverse
from wbcore import serializers
from wbcore.contrib.authentication.models import User
from wbcore.contrib.authentication.serializers import UserRepresentationSerializer
from wbcore.utils.urls import new_mode
from wbcrm.models import Account
from wbcrm.serializers.accounts import AccountRepresentationSerializer

from wbportfolio.models import AccountReconciliation
from wbportfolio.models.reconciliations import AccountReconciliationLine
from wbportfolio.serializers.products import ProductRepresentationSerializer

if TYPE_CHECKING:
    from rest_framework.request import Request


class AccountReconciliationModelSerializer(serializers.ModelSerializer):
    creator = serializers.PrimaryKeyRelatedField(required=False, queryset=User.objects.all())
    _creator = UserRepresentationSerializer(source="creator")
    account = serializers.PrimaryKeyRelatedField(
        default=serializers.DefaultFromGET("account"), queryset=Account.objects.all()
    )
    _account = AccountRepresentationSerializer(source="account")
    approved_by = serializers.PrimaryKeyRelatedField(required=False, queryset=User.objects.all())
    _approved_by = UserRepresentationSerializer(source="approved_by")

    def validate(self, data: dict) -> dict:
        data["creator"] = self.context["request"].user
        if (account := data.get("account", None)) and (isinstance(account, str) or isinstance(account, list)):
            account = account[0] if isinstance(account, list) else account
            data["account"] = Account.objects.get(id=account)
        return data

    @serializers.register_resource()
    def register_lines(self, instance: AccountReconciliation, request: "Request", user: "User") -> dict[str, str]:
        return {
            "lines": reverse(
                "wbportfolio:accountreconciliation-accountreconciliationline-list", args=[instance.id], request=request
            )
        }

    @serializers.register_resource()
    def register_reconcile(self, instance: AccountReconciliation, request: "Request", user: "User") -> dict[str, str]:
        if instance.approved_dt:
            return {}

        if not user.is_internal:
            if hasattr(instance, "unequal_exists") and instance.unequal_exists:
                return {
                    "disagree_customer": reverse(
                        "wbportfolio:accountreconciliation-disagree-customer", args=[instance.id], request=request
                    )
                }
            elif hasattr(instance, "unequal_exists") and not instance.unequal_exists:
                return {
                    "agree_customer": reverse(
                        "wbportfolio:accountreconciliation-agree-customer", args=[instance.id], request=request
                    )
                }

        return {}

    @serializers.register_resource()
    def register_claims(self, instance: AccountReconciliation, request: "Request", user: "User") -> dict[str, str]:
        return {
            "claims": reverse("wbportfolio:account-claim-list", args=[instance.account.id], request=request),
            "add-claim": new_mode(reverse("wbportfolio:claim-list", request=request)),
        }

    @serializers.register_resource()
    def register_recompute(self, instance: AccountReconciliation, request: "Request", user: "User") -> dict[str, str]:
        if instance.approved_dt:
            return {}

        if user.is_internal:
            return {
                "recalculate": reverse(
                    "wbportfolio:accountreconciliation-recompute", args=[instance.id], request=request
                ),
                "notify": reverse("wbportfolio:accountreconciliation-notify", args=[instance.id], request=request),
            }

        return {}

    class Meta:
        model = AccountReconciliation
        fields = (
            "id",
            "reconciliation_date",
            "creator",
            "_creator",
            "approved_by",
            "_approved_by",
            "approved_dt",
            "account",
            "_account",
            "_additional_resources",
        )


class AccountReconciliationLineModelSerializer(serializers.ModelSerializer):
    _product = ProductRepresentationSerializer(source="product")
    is_equal = serializers.BooleanField(read_only=True)

    assets_under_management = serializers.IntegerField(
        read_only=True,
        decorators=[serializers.decorator(position="left", value="{{currency_key}}", decorator_type="text")],
    )
    assets_under_management_external = serializers.IntegerField(
        read_only=True,
        decorators=[serializers.decorator(position="left", value="{{currency_key}}", decorator_type="text")],
    )
    assets_under_management_diff = serializers.IntegerField(
        read_only=True,
        decorators=[serializers.decorator(position="left", value="{{currency_key}}", decorator_type="text")],
    )
    shares_diff = serializers.IntegerField(read_only=True)
    nominal_value_diff = serializers.IntegerField(read_only=True)
    pct_diff = serializers.DecimalField(read_only=True, max_digits=5, decimal_places=4)

    price = serializers.DecimalField(read_only=True, max_digits=18, decimal_places=2, label="Price")

    currency = serializers.CharField(required=False)
    currency_key = serializers.CharField(required=False)

    @serializers.register_resource()
    def register_claims(self, instance: AccountReconciliationLine, request: "Request", user: "User") -> dict[str, str]:
        return {
            "product-claims": f'{reverse("wbportfolio:account-claim-list", args=[instance.reconciliation.account.id], request=request)}?product={instance.product.id}',
            "add-product-claim": new_mode(
                f'{reverse("wbportfolio:claim-list", request=request)}?product={instance.product.id}'
            ),
        }

    def validate(self, data: dict[str, Any]) -> dict[str, Any]:
        # If we get either external shares or external nominal value, we have to set the other one explicitly to None
        # to be sure that the other side is correctly computed inside the save method
        if "shares_external" not in data and "nominal_value_external" in data:
            data["shares_external"] = None

        if "shares_external" in data and "nominal_value_external" not in data:
            data["nominal_value_external"] = None

        return super().validate(data)

    class Meta:
        model = AccountReconciliationLine
        read_only_fields = ("product", "shares", "nominal_value")
        decorators = {"price": serializers.decorator(position="left", value="{{currency_key}}", decorator_type="text")}
        percent_fields = ["pct_diff"]
        fields = (
            "id",
            "reconciliation",
            "price",
            "price_date",
            "currency",
            "currency_key",
            "product",
            "_product",
            "shares",
            "nominal_value",
            "assets_under_management",
            "shares_external",
            "nominal_value_external",
            "assets_under_management_external",
            "shares_diff",
            "nominal_value_diff",
            "assets_under_management_diff",
            "pct_diff",
            "is_equal",
            "_additional_resources",
        )



from wbcore.menus import ItemPermission, MenuItem

PRODUCTPERFORMANCELIST_MENUITEM = MenuItem(
    label="Performances",
    endpoint="wbportfolio:productperformance-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbportfolio.view_product"]
    ),
)

PRODUCTPERFORMANCECOMPARISONLIST_MENUITEM = MenuItem(
    label="Performances Summary",
    endpoint="wbportfolio:productperformancecomparison-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbportfolio.view_product"]
    ),
)
PRODUCTPERFORMANCENNMLIST_MENUITEM = MenuItem(
    label="Net new money",
    endpoint="wbportfolio:productperformancennmlist-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbportfolio.view_product"]
    ),
)
